import time

import numpy as np

from goofi.data import DataType
from goofi.node import Node
from goofi.params import BoolParam, FloatParam, StringParam


class ERP(Node):
    def config_input_slots():
        return {"signal": DataType.ARRAY, "trigger": DataType.ARRAY}

    def config_output_slots():
        return {"erp": DataType.ARRAY}

    def config_params():
        return {
            "erp": {
                "duration": FloatParam(1.0, vmin=0.0, vmax=3.0, doc="Duration of the ERP in seconds"),
                "baseline": StringParam("mean", options=["none", "mean"], doc="Baseline correction method"),
                "baseline_axis": -1,
                "reset": BoolParam(False, trigger=True, doc="Reset average evoked response"),
                "delay_retrigger": BoolParam(True, doc="Wait for length of signal before retriggering ERP"),
            }
        }

    def setup(self):
        self.collecting = None
        self.erp = None
        self.n_samples = 0
        self.last_trigger = None

    def process(self, signal, trigger):
        if self.params.erp.reset.value:
            self.erp = None
            self.collecting = None

        if signal is None:
            return None

        triggered = trigger is not None and not np.any(np.isnan(trigger.data))
        if triggered and self.params.erp.delay_retrigger.value and self.collecting is not None:
            # ignore trigger as we are still collecting data for the last trigger
            triggered = False

        if triggered:
            # new trigger, start collecting data
            self.collecting = []
        elif self.collecting is None:
            # no trigger, not collecting data
            return None

        self.collecting.append(signal.data)

        target_samples = self.params.erp.duration.value * signal.meta["sfreq"]
        if sum(chunk.shape[self.params.erp.baseline_axis.value] for chunk in self.collecting) >= target_samples:
            # enough samples collected, add to ERP
            data = np.concatenate(self.collecting, axis=self.params.erp.baseline_axis.value)
            data = np.take(data, np.arange(target_samples, dtype=int), axis=self.params.erp.baseline_axis.value)

            # baseline correction
            if self.params.erp.baseline.value == "mean":
                data = data - data.mean(axis=self.params.erp.baseline_axis.value, keepdims=True)

            if self.erp is None:
                # first ERP
                self.erp = data
                self.n_samples = 1
            elif self.erp.shape != data.shape:
                raise ValueError(f"Data shape ({data.shape}) changed, ERP has shape {self.erp.shape}.")
            else:
                # add new ERP to average
                self.erp = self.erp + data
                self.n_samples += 1

            self.last_trigger = time.time()
            self.collecting = None

            meta = signal.meta.copy()
            meta["erp_samples"] = self.n_samples
            return {"erp": (self.erp / self.n_samples, meta)}
        else:
            return None

    def erp_duration_changed(self):
        self.erp = None
        self.collecting = None



import cv2
import numpy as np
from PIL import Image

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, StringParam


class Embedding(Node):

    def config_input_slots():
        return {"text": DataType.STRING, "data": DataType.ARRAY}

    def config_output_slots():
        return {
            "text_embeddings": DataType.ARRAY,
            "data_embeddings": DataType.ARRAY,
        }

    def config_params():
        return {
            "embedding": {
                "model": StringParam(
                    "openai/clip-vit-base-patch32",
                    options=[
                        "openai/clip-vit-base-patch32",
                        "openai/clip-vit-large-patch14",
                        "laion/CLIP-ViT-H-14-laion2B-s32B-b79K",
                        "all-MiniLM-L6-v2",
                        "word2vec",
                    ],
                    doc="Model ID or name for embedding generation",
                ),
                "split_by_comma": BoolParam(
                    False,
                    doc="Whether to split text input by comma and generate embeddings for each part separately",
                ),
                "device": StringParam("cpu", options=["cpu", "cuda"]),
            }
        }

    def setup(self):
        # Explicitly reset attributes
        self.model = None
        self.processor = None
        self.model_type = None
        self.device = None

        # Select device
        import torch

        self.torch = torch

        torch.set_grad_enabled(False)

        self.device = torch.device(self.params.embedding.device.value)
        self.model_id = self.params.embedding.model.value
        print(f"Selected model: {self.model_id}, using device: {self.device}")

        # Load CLIP models
        if "clip" in self.model_id.lower():
            self.model_type = "clip"

            print("Initializing CLIP model...")
            from transformers import CLIPModel, CLIPProcessor

            try:
                # try loading the local model (local_files_only=False interferes with goofi-pipe's multiprocessing environment)
                # related issue: https://github.com/CompVis/stable-diffusion/issues/90#issuecomment-1228726914
                self.model = CLIPModel.from_pretrained(self.model_id, local_files_only=True).to(self.device)
            except OSError:
                self.model = CLIPModel.from_pretrained(self.model_id).to(self.device)

            self.processor = CLIPProcessor.from_pretrained(self.model_id)
        # Load other models
        elif "sbert" in self.model_id.lower() or "MiniLM" in self.model_id:
            self.model_type = "sbert"

            print("Initializing SBERT model...")
            from sentence_transformers import SentenceTransformer

            self.model = SentenceTransformer(self.model_id).to(self.device)
        elif "fasttext" in self.model_id.lower():
            self.model_type = "fasttext"

            print("Initializing FastText model...")
            import gensim.downloader as api

            self.model = api.load("fasttext-wiki-news-subwords-300")
        elif "word2vec" in self.model_id.lower():
            self.model_type = "word2vec"

            print("Initializing Word2Vec model...")
            import gensim.downloader as api

            self.model = api.load("word2vec-google-news-300")
        else:
            raise ValueError(f"Unsupported model type: {self.model_id}")
        print(f"Model {self.model_id} initialized successfully as {self.model_type} on {self.device}.")

    def process(self, text: Data, data: Data):
        # Initialize outputs and metadata
        text_embeddings = None  # Default to None
        data_embeddings = None  # Default to None
        metadata = {}

        # Compute text embeddings if text input is provided
        if text is not None:
            input_text = text.data
            metadata["text"] = text.meta

            if self.params.embedding.split_by_comma.value:
                input_texts = [t.strip() for t in input_text.split(",")]
            else:
                input_texts = [input_text]

            with self.torch.inference_mode():
                if self.model_type == "clip":  # Use CLIP for text
                    inputs_text = self.processor(text=input_texts, return_tensors="pt", padding=True).to(self.device)
                    outputs_text = self.model.get_text_features(**inputs_text)
                    text_embeddings = outputs_text.cpu().numpy()

                elif self.model_type == "sbert":  # Use SBERT for text
                    text_embeddings = self.model.encode(input_texts, convert_to_numpy=True)

                elif self.model_type == "fasttext":  # Use FastText for text
                    text_embeddings = []
                    word_vectors = {
                        word: self.model[word] for sentence in input_texts for word in sentence.split() if word in self.model
                    }
                    for sentence in input_texts:
                        words = sentence.split()
                        vectors = [word_vectors[word] for word in words if word in word_vectors]
                        if vectors:
                            sentence_embedding = np.mean(vectors, axis=0)
                            text_embeddings.append(sentence_embedding)
                    text_embeddings = np.array(text_embeddings) if text_embeddings else None

                elif self.model_type == "word2vec":  # Use Word2Vec for text
                    text_embeddings = []
                    for sentence in input_texts:
                        words = sentence.split()
                        vectors = [self.model[word] for word in words if word in self.model]
                        if vectors:
                            sentence_embedding = np.mean(vectors, axis=0)
                            text_embeddings.append(sentence_embedding)
                    text_embeddings = np.array(text_embeddings) if text_embeddings else None

                else:
                    raise ValueError(f"Model {self.model_id} does not support text embeddings.")

        # Compute image embeddings if data input is provided
        if data is not None:
            input_data = data.data
            metadata["data"] = data.meta  # Preserve metadata

            with self.torch.inference_mode():
                # Preprocess and generate image embeddings using CLIP
                if self.model_type == "clip":
                    # Check and preprocess numpy array
                    if input_data.ndim == 4 and input_data.shape[0] == 1:  # Remove batch dimension if present
                        input_data = input_data.squeeze(0)

                    # Ensure image is resized and normalized
                    if input_data.shape[:2] != (224, 224):  # Resize if not already 224x224
                        input_data = cv2.resize(input_data, (224, 224))

                    # Normalize pixel values if necessary
                    if input_data.max() > 1.0:  # Normalize to [0, 1] if data is in [0, 255]
                        input_data = np.clip(input_data / 255.0, 0.0, 1.0)

                    # Convert to PIL Image for CLIP compatibility
                    input_data_pil = Image.fromarray((input_data * 255).astype(np.uint8))

                    # Use CLIPProcessor for final preprocessing
                    inputs_data = self.processor(images=input_data_pil, return_tensors="pt", padding=True)
                    inputs_data = {k: v.to(self.device) for k, v in inputs_data.items()}

                    # Generate embeddings
                    outputs_data = self.model.get_image_features(**inputs_data)
                    data_embeddings = outputs_data.cpu().numpy()
                else:
                    raise ValueError(f"Model {self.model_id} does not support image embeddings.")

        # Return separate embeddings and a dictionary of metadata
        return {
            "text_embeddings": (text_embeddings, metadata) if text_embeddings is not None else None,
            "data_embeddings": (data_embeddings, metadata) if data_embeddings is not None else None,
        }

    def embedding_model_changed(self, _):
        """Reinitialize the stream."""
        self.setup()

    def embedding_device_changed(self, _):
        """Reinitialize the stream."""
        self.setup()



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, FloatParam, IntParam, StringParam


class Buffer(Node):
    def config_input_slots():
        return {"val": DataType.ARRAY}

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def config_params():
        return {
            "buffer": {
                "size": FloatParam(10, 1, 5000, doc="Buffer size in samples or seconds, depending on the unit."),
                "axis": IntParam(-1, doc="Axis along which to buffer the data. Negative values count from the end."),
                "unit": StringParam(
                    "samples",
                    options=["samples", "seconds"],
                    doc=(
                        "Unit of the buffer size. If 'seconds', the metadata must specify "
                        "the sampling frequency via the 'sfreq' key."
                    ),
                ),
                "reset": BoolParam(False, trigger=True, doc="Clear the buffer"),
            }
        }

    def setup(self):
        self.name_buffer = None
        self.buffer = None

    def process(self, val: Data):
        if val is None:
            return None

        if self.params.buffer.reset.value:
            # reset buffer
            self.buffer = None

        unit = self.params.buffer.unit.value
        if unit == "samples":
            maxlen = int(self.params.buffer.size.value)
        elif unit == "seconds":
            # get sampling frequency from metadata
            if "sfreq" not in val.meta:
                raise ValueError("If unit is 'seconds', the metadata must contain 'sfreq'.")
            maxlen = int(self.params.buffer.size.value * val.meta["sfreq"])
        else:
            raise ValueError(f"Unknown unit: {unit}")

        axis = self.params.buffer.axis.value
        # convert negative axis to positive
        if axis < 0:
            axis = val.data.ndim + axis

        if self.buffer is not None:
            # add dimension to buffer and val.data if axis is out of bounds
            if self.buffer.ndim < (axis + 1):
                self.buffer = np.expand_dims(self.buffer, axis=axis)
            if val.data.ndim < (axis + 1):
                val.data = np.expand_dims(val.data, axis=axis)

            try:
                # concatenate data to buffer
                self.buffer = np.concatenate((self.buffer, val.data), axis=axis)
                # update channel names
                if f"dim{axis}" in val.meta["channels"]:
                    self.name_buffer += val.meta["channels"][f"dim{axis}"]
            except ValueError:
                # data shape changed, reset buffer
                self.buffer = None

        # (re-)initialize buffer
        if self.buffer is None:
            self.buffer = np.array(val.data)
            self.name_buffer = val.meta["channels"][f"dim{axis}"] if f"dim{axis}" in val.meta["channels"] else None

        # remove data that exceeds the maximum length
        if self.buffer.shape[axis] > maxlen:
            self.buffer = np.take(self.buffer, range(-maxlen, 0), axis=axis)
            if self.name_buffer is not None:
                self.name_buffer = self.name_buffer[-maxlen:]

        # update channel names
        if self.name_buffer is not None:
            val.meta["channels"][f"dim{axis}"] = self.name_buffer

        return {"out": (self.buffer, val.meta)}



from goofi.params import StringParam
from goofi.data import Data, DataType
from goofi.node import Node
import numpy as np


class PowerBandEEG(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {
            "delta": DataType.ARRAY,
            "theta": DataType.ARRAY,
            "alpha": DataType.ARRAY,
            "lowbeta": DataType.ARRAY,
            "highbeta": DataType.ARRAY,
            "gamma": DataType.ARRAY,
        }

    def config_params():
        return {
            "powerband": {
                "power_type": StringParam("absolute", options=["absolute", "relative"]),
            }
        }

    def process(self, data: Data):
        if data is None or data.data is None:
            return None
        bands = {
            "delta": (1, 3),
            "theta": (3, 7),
            "alpha": (7, 12),
            "lowbeta": (12, 20),
            "highbeta": (20, 30),
            "gamma": (30, 50),
        }
        power_type = self.params["powerband"]["power_type"].value
        if data.data.ndim == 1:
            freqs = np.array(data.meta["channels"]["dim0"])
            if freqs[0] == 0:
                freqs[0] = 1e-8
            del data.meta["channels"]["dim0"]
        elif data.data.ndim == 2:
            freqs = np.array(data.meta["channels"]["dim1"])
            if freqs[0] == 0:
                freqs[0] = 1e-8
            del data.meta["channels"]["dim1"]

        output = {}
        for band, (f_min, f_max) in bands.items():
            valid_indices = np.where((freqs >= f_min) & (freqs <= f_max))[0]
            if data.data.ndim == 1:
                selected_psd = data.data[valid_indices]
            else:  # if 2D
                selected_psd = data.data[:, valid_indices]

            # Computing the power
            power = np.sum(selected_psd, axis=-1)
            if power_type == "relative":
                total_power = np.sum(data.data, axis=-1)
                power = power / total_power

            output[band] = (np.array(power), {"freq_min": f_min, "freq_max": f_max, **data.meta})
        return output



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, IntParam, StringParam


class Reservoir(Node):
    def config_params():
        return {
            "reservoir": {
                "size": IntParam(100, 2, 1000),
                "function": StringParam("tanh", options=["tanh", "sigmoid"]),
                "add_bias": True,
                "mean": FloatParam(0.0, -0.2, 0.2),
                "std": FloatParam(1.0, 0.7, 1.3),
            },
            "common": {"autotrigger": True},
        }

    def config_input_slots():
        return {"connectivity": DataType.ARRAY}

    def config_output_slots():
        return {"data": DataType.ARRAY}

    def setup(self):
        self.reservoir = np.zeros(self.params.reservoir.size.value)
        self.weights = np.random.uniform(-1, 1, (self.params.reservoir.size.value, self.params.reservoir.size.value))
        self.bias = np.random.uniform(-1, 1, self.params.reservoir.size.value)

    def process(self, connectivity: Data):
        if connectivity is None:
            w = self.weights
        else:
            assert connectivity.data.ndim == 2, "Connectivity matrix must be two-dimensional."
            assert connectivity.data.shape[0] == connectivity.data.shape[1], "Connectivity matrix must be square."
            assert (
                connectivity.data.shape[0] == self.params.reservoir.size.value
            ), "Connectivity matrix must have the same size as the reservoir."
            conn = connectivity.data
            conn[np.diag_indices_from(conn)] = 0
            w = conn

        # scale the weights
        w = w * self.params.reservoir.std.value + self.params.reservoir.mean.value

        self.reservoir = np.dot(w, self.reservoir)
        if self.params.reservoir.add_bias.value:
            self.reservoir += self.bias

        if self.params.reservoir.function.value == "tanh":
            self.reservoir = np.tanh(self.reservoir)
        elif self.params.reservoir.function.value == "sigmoid":
            self.reservoir = 1 / (1 + np.exp(-self.reservoir))
        else:
            raise ValueError(f"Unsupported function {self.params.reservoir.function.value} for reservoir output.")

        # reshape into matrix of size greatest common divisor of reservoir size
        return {"data": (self.reservoir, {"sfreq": self.params.common.max_frequency.value})}

    def reservoir_size_changed(self, value):
        self.setup()



from goofi.data import Data, DataType
from goofi.node import Node


class SetMeta(Node):
    def config_input_slots():
        return {"array": DataType.ARRAY}

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def config_params():
        return {"meta": {"key": "key", "value": "value", "type": "float"}}

    def process(self, array: Data):
        if array is None:
            return None

        val = self.params.meta.value.value

        if self.params.meta.type.value == "int":
            val = int(val)
        elif self.params.meta.type.value == "float":
            val = float(val)
        elif self.params.meta.type.value == "bool":
            val = bool(val)
        elif self.params.meta.type.value == "str":
            val = str(val)
        else:
            raise ValueError(f"Unknown type {self.params.meta.type.value}")

        array.meta[self.params.meta.key.value] = val
        return {"out": (array.data, array.meta)}



from goofi.data import DataType
from goofi.node import Node
from goofi.params import StringParam


class ConstantString(Node):
    def config_params():
        return {
            "constant": {"value": StringParam("default_value")},
            "common": {"autotrigger": True},
        }

    def config_output_slots():
        return {"out": DataType.STRING}

    def process(self):
        return {"out": (self.params["constant"]["value"].value, {})}



from copy import deepcopy

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import IntParam, StringParam


class HarmonicSpectrum(Node):
    def config_input_slots():
        return {"psd": DataType.ARRAY}

    def config_output_slots():
        return {"harmonic_spectrum": DataType.ARRAY, "max_harmonicity": DataType.ARRAY, "avg_harmonicity": DataType.ARRAY}

    def config_params():
        return {
            "harmonic": {
                "metric": StringParam("harmsim", options=["harmsim", "subharm_tension"]),
                "normalize": True,
                "power_law_remove": True,
                "n_harms": IntParam(3, 1, 10, doc="Number of harmonics to consider in the subharmonic tension metric"),
                "delta_lim": IntParam(250, 1, 300, doc="Maximum delta in the subharmonic tension metric"),
                "min_notes": IntParam(2, 1, 10, doc="Minimum number of notes in the subharmonic tension metric"),
            }
        }

    def setup(self):
        from biotuner.biotuner_utils import apply_power_law_remove
        from biotuner.metrics import compute_subharmonic_tension, dyad_similarity

        self.apply_power_law_remove = apply_power_law_remove
        self.compute_subharmonic_tension = compute_subharmonic_tension
        self.dyad_similarity = dyad_similarity

        self.cached_freqs = None
        self.cached_harmonicity_matrix = None
        self.cached_metric = None

    def process(self, psd: Data):
        if psd is None or psd.data is None:
            return None
        psd.data = np.squeeze(psd.data)
        metadata = psd.meta

        if psd.data.ndim == 1:
            freqs = metadata["channels"]["dim0"]
            if freqs[0] == 0:
                freqs[0] = 1e-8
            harmonicity_values = self._compute_for_single_psd(psd.data, freqs)
            updated_metadata = {**metadata, "channels": {"dim0": list(range(len(harmonicity_values)))}}

            reduced_meta = deepcopy(updated_metadata)
            if "channels" in reduced_meta:
                del reduced_meta["channels"]

            return {
                "harmonic_spectrum": (harmonicity_values, updated_metadata),
                "max_harmonicity": (np.max(harmonicity_values), {"freq": freqs, **reduced_meta}),
                "avg_harmonicity": (np.mean(harmonicity_values), {"freq": freqs, **reduced_meta}),
            }
        elif psd.data.ndim == 2:
            freqs = metadata["channels"]["dim1"]
            if freqs[0] == 0:
                freqs[0] == 1e-8
            harmonicity_values_matrix = np.zeros(psd.data.shape)
            for i, row in enumerate(psd.data):
                harmonicity_values_matrix[i, :] = self._compute_for_single_psd(row, freqs)

            reduced_meta = deepcopy(metadata)
            if "channels" in reduced_meta and "dim1" in reduced_meta["channels"]:
                del reduced_meta["channels"]["dim1"]

            return {
                "harmonic_spectrum": (harmonicity_values_matrix, {**metadata}),
                "max_harmonicity": (np.max(harmonicity_values_matrix, axis=1), {"freq": freqs, **reduced_meta}),
                "avg_harmonicity": (np.mean(harmonicity_values_matrix, axis=1), {"freq": freqs, **reduced_meta}),
            }
        else:
            raise ValueError("Data must be either 1D or 2D")

    def _compute_for_single_psd(self, psd_values, freqs):
        power_law_remove = self.params["harmonic"]["power_law_remove"].value
        if power_law_remove:
            psd_values = self.apply_power_law_remove(freqs, psd_values, power_law_remove)
            # clip the psd_values to avoid negative values by converting them to 0
            psd_values = np.clip(psd_values, 0, None)

        metric = self.params["harmonic"]["metric"].value
        normalize = self.params["harmonic"]["normalize"].value
        n_harms = self.params["harmonic"]["n_harms"].value
        delta_lim = self.params["harmonic"]["delta_lim"].value
        min_notes = self.params["harmonic"]["min_notes"].value

        if np.array_equal(self.cached_freqs, freqs) and self.cached_metric == metric:
            harmonicity_matrix = self.cached_harmonicity_matrix
        else:
            harmonicity_matrix = self.compute_harmonicity(freqs, metric, n_harms, delta_lim, min_notes)
            # Cache the freqs and harmonicity matrix for next time
            self.cached_freqs = freqs
            self.cached_harmonicity_matrix = harmonicity_matrix
            self.cached_metric = metric  # Cache the current metric
        harmonicity_values = np.zeros(len(freqs))

        total_power = np.sum(psd_values**2)

        for i in range(len(freqs)):
            weighted_sum_harmonicity = 0
            for j in range(len(freqs)):
                if i != j:
                    weighted_sum_harmonicity += harmonicity_matrix[i, j] * (psd_values[i] * psd_values[j])

            harmonicity_values[i] = (weighted_sum_harmonicity / (2 * total_power)) if normalize else weighted_sum_harmonicity

        return harmonicity_values

    def compute_harmonicity(self, freqs, metric, n_harms, delta_lim, min_notes):
        harmonicity = np.zeros((len(freqs), len(freqs)))

        for i, f1 in enumerate(freqs):
            for j, f2 in enumerate(freqs):
                if f1 != f2:  # avoiding division by zero and self-comparison
                    if metric == "harmsim":
                        harmonicity[i, j] = self.dyad_similarity(f1 / f2)
                    elif metric == "subharm_tension":
                        _, _, subharm, _ = self.compute_subharmonic_tension(
                            [f1, f2], n_harmonics=n_harms, delta_lim=delta_lim, min_notes=min_notes
                        )
                        try:
                            harmonicity[i, j] = 1 - subharm[0]
                        except Exception:
                            harmonicity[i, j] = None
        return harmonicity



from collections import defaultdict

import numpy as np

from goofi.data import DataType
from goofi.node import Node
from goofi.params import BoolParam


class ProbabilityMatrix(Node):

    def config_input_slots():
        return {"input_data": DataType.ARRAY}

    def config_params():
        return {
            "Probability": {
                "Reset": BoolParam(trigger=True),
            }
        }

    def config_output_slots():
        return {"data": DataType.ARRAY}

    def setup(self):
        self.state_to_idx = {}  # Mapping of state to index
        self.idx_to_state = []  # List of states, where index corresponds to state's index
        self.transition_count = defaultdict(lambda: defaultdict(int))
        self.transition_matrix = None
        self.last_state = None

    def update_transition_matrix(self):
        """Update the transition matrix based on transition counts."""
        num_states = len(self.idx_to_state)
        self.transition_matrix = np.zeros((num_states, num_states))
        for i, targets in self.transition_count.items():
            total_transitions = sum(targets.values())
            for j, count in targets.items():
                self.transition_matrix[i][j] = count / total_transitions

    def process(self, input_data):
        if input_data is None:
            return None

        # Discretize the input data by rounding to 2 decimal places
        data = [round(val, 2) for val in input_data.data]

        # Update state-to-index mapping and index-to-state list
        for state in data:
            if state not in self.state_to_idx:
                self.state_to_idx[state] = len(self.idx_to_state)
                self.idx_to_state.append(state)

        # Handle sequential data
        for current_state in data:
            if self.last_state is not None:
                i = self.state_to_idx[self.last_state]
                j = self.state_to_idx[current_state]
                self.transition_count[i][j] += 1
            self.last_state = current_state

        # Update the transition matrix
        self.update_transition_matrix()

        if self.params.Probability.Reset.value:
            self.setup()
        return {"data": (self.transition_matrix, {"sfreq": self.params.common.autotrigger.value})}

    def autotrigger_changed(self, value):
        self.setup()



from goofi.params import IntParam, FloatParam
from goofi.data import Data, DataType
from goofi.node import Node
import numpy as np
import time


class FractalImage(Node):
    def config_input_slots():
        return {"complexity": DataType.ARRAY}  # No input data for this node

    def config_output_slots():
        return {"image": DataType.ARRAY}

    def config_params():
        return {
            "image": {
                "image_size": IntParam(512, 128, 2048),
                "persistence": FloatParam(
                    0.5, 0.1, 1.0, doc="The persistence value for the fractal noise (higher values result in more detail)"
                ),
                "octaves": IntParam(6, 1, 6),
                "lacunarity": FloatParam(2.0, 1.0, 4.0),
            },
            "common": {"autotrigger": True},
        }

    def process(self, complexity: Data):
        # Extract the parameters
        image_size = self.params["image"]["image_size"].value
        octaves = self.params["image"]["octaves"].value
        lacunarity = self.params["image"]["lacunarity"].value
        if complexity is None:
            persistence = self.params["image"]["persistence"].value
        else:
            persistence = complexity.data

        # Generate a new seed based on current time
        seed = int(time.time() * 1000) % 2**32  # Convert time to milliseconds and take modulo to avoid overflow

        # Generate the fractal image
        x = np.linspace(0, 5, image_size)
        y = np.linspace(0, 5, image_size)
        x, y = np.meshgrid(x, y)
        image = fbm(x, y, octaves=octaves, lacunarity=lacunarity, persistence=persistence, seed=seed)
        image = (image + 1) / 2  # Rescale values from [-1, 1] to [0, 1]
        # Return the generated fractal image
        return {"image": (image, {})}  # No metadata to include, hence the empty dictionary


def fade(t):
    return 6 * t**5 - 15 * t**4 + 10 * t**3


def lerp(a, b, x):
    return a + x * (b - a)


def gradient(h, x, y):
    # Define the gradient vectors
    vectors = np.array([[0, 1], [0, -1], [1, 0], [-1, 0]])
    # Choose the appropriate gradient vector based on hash value h
    g = vectors[h % 4]
    # Calculate the dot product
    return g[..., 0] * x + g[..., 1] * y


def perlin(x, y, seed=0):
    np.random.seed(seed)
    p = np.arange(256, dtype=int)
    np.random.shuffle(p)
    p = np.stack([p, p]).flatten()
    xi = x.astype(int)
    yi = y.astype(int)
    xf = x - xi
    yf = y - yi

    u = fade(xf)
    v = fade(yf)

    n00 = gradient(p[p[xi] + yi], xf, yf)
    n01 = gradient(p[p[xi] + yi + 1], xf, yf - 1)
    n11 = gradient(p[p[xi + 1] + yi + 1], xf - 1, yf - 1)
    n10 = gradient(p[p[xi + 1] + yi], xf - 1, yf)

    x1 = lerp(n00, n10, u)
    x2 = lerp(n01, n11, u)

    return lerp(x1, x2, v)


def fbm(x, y, octaves=6, lacunarity=2, persistence=0.5, seed=0):
    value = 0
    amplitude = 1.0
    frequency = 1.0
    for _ in range(octaves):
        value += amplitude * perlin(x * frequency, y * frequency, seed=seed)
        amplitude *= persistence
        frequency *= lacunarity
    return value



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional

import numpy as np


class DataType(Enum):
    """
    The type of data contained in a data object. The data type determines the type of the data field in the data
    object.

    - `ARRAY`: An n-dimensional numpy array.
    - `STRING`: A string.
    - `TABLE`: A named map of data objects (i.e. a dict of str keys and Data values).
    """

    ARRAY = 0
    STRING = 1
    TABLE = 2

    def empty(self) -> Any:
        """
        Get an empty numpy array of the correct type for this data type.
        """
        if self == DataType.ARRAY:
            return np.zeros(0)
        elif self == DataType.STRING:
            return ""
        elif self == DataType.TABLE:
            return dict()
        else:
            raise ValueError(f"Unknown data type {self}")

    def __str__(self) -> str:
        return self.name


@dataclass
class Data:
    """
    Data objects are used to send data between nodes. They contain the data object (`data`) itself, the data type
    (`dtype`), and some metadata (`meta`). The data types are limited to the types defined in the `DataType` enum.
    The data field must be of the correct type for the data type. The metadata field is a dict that can contain
    any metadata about the data. The metadata field can be used to store information about the data, such as
    channel names, sampling frequencies, etc.

    ### Parameters
    `dtype` : DataType
        The data type of the data object.
    `data` : Any
        A data object matching the data type.
    `meta` : Dict[str, Any]
        The metadata dictionary.
    """

    dtype: DataType
    data: Any
    meta: Dict[str, Any]

    def __post_init__(self):
        """
        Check that the data type is valid. The data field must be of the correct type for the data type.
        """
        # general checks
        if self.dtype is None or not isinstance(self.dtype, DataType):
            raise ValueError(f"Expected dtype of type DataType, got {type(self.dtype)}.")
        if self.meta is None or not isinstance(self.meta, dict):
            raise ValueError(f"Expected meta of type dict, got {type(self.meta)}.")

        # validate data type
        if isinstance(self.dtype, DataType) and self.dtype not in DTYPE_TO_TYPE:
            raise RuntimeError(
                f"Data type {self.dtype} is defined but not handled in check_data. "
                "Please report this bug at https://github.com/dav0dea/goofi-pipe/issues."
            )
        if not isinstance(self.data, DTYPE_TO_TYPE[self.dtype]):
            raise ValueError(f"Expected data of type {DTYPE_TO_TYPE[self.dtype]}, got {type(self.data)}.")

        # configure the data object by calling the appropriate configure method
        getattr(self, f"_configure_{self.dtype.name.lower()}")()

    def _configure_array(self):
        """
        Configure the data object as an array and populate the metadata.
        """
        # make sure that arrays are at least 1-dimensional
        if self.data.ndim == 0:
            self.data = np.array([self.data])

        # populate the metadata
        self.meta["shape"] = self.data.shape

        # make sure that the metadata contains a channels dict
        if "channels" in self.meta:
            assert isinstance(self.meta["channels"], dict), "Expected channels to be a dict."
        else:
            self.meta["channels"] = {}

        # check the channels metadata
        for dim in range(self.data.ndim):
            if f"dim{dim}" in self.meta["channels"]:
                assert isinstance(self.meta["channels"][f"dim{dim}"], list), f"Expected dim{dim} to be a list."
                assert len(self.meta["channels"][f"dim{dim}"]) == self.data.shape[dim], (
                    f"Expected dim{dim} to have length {self.data.shape[dim]} but got "
                    f"{len(self.meta['channels'][f'dim{dim}'])}."
                )
        for dim in self.meta["channels"].keys():
            assert dim.startswith("dim"), f"Expected channel key to start with 'dim', got {dim}."
            dim = dim[3:]
            assert dim.isdigit(), f"Expected channel key to end with a number, got {dim}."
            dim = int(dim)
            assert dim < self.data.ndim, f"Expected channel key to be less than {self.data.ndim}, got {dim}."

    def _configure_string(self):
        pass

    def _configure_table(self):
        """
        Configure the data object as a table and populate the metadata.
        """
        for key, value in self.data.items():
            # make sure that table keys are strings
            if not isinstance(key, str):
                raise ValueError(f"Expected table keys of type str, got {type(key)}.")
            # make sure that table values are Data objects
            if not isinstance(value, Data):
                raise ValueError(f"Expected table values of type Data, got {type(value)}.")


def to_data(obj: Any, meta: Optional[Dict[str, Any]] = None):
    """
    Convert an object to a Data object. The data type of the Data object is determined by the type of the object.
    The object must be of a supported type (i.e. numpy array, string, or dict).

    ### Parameters
    `obj` : Any
        The object to convert to a Data object.
    `meta` : Dict[str, Any], optional
        The metadata dictionary.

    ### Returns
    Data
        The converted Data object.
    """
    for dtype, types in DTYPE_TO_TYPE.items():
        if isinstance(obj, types):
            return Data(dtype=dtype, data=obj, meta=meta or {})
    raise ValueError(f"Could not parse goofi dtype for object of type {type(obj)}.")


DTYPE_TO_TYPE = {
    DataType.ARRAY: (np.ndarray, np.number),
    DataType.STRING: str,
    DataType.TABLE: dict,
}



import ast

import numpy as np

from goofi.data import DataType, to_data
from goofi.node import Node
from goofi.params import BoolParam, StringParam


def convert_to_numpy(val):
    try:
        parsed_val = ast.literal_eval(val)  # Convert string to Python list/dict
        if isinstance(parsed_val, list):
            return np.array(parsed_val, dtype=np.float32)  # Convert lists to NumPy arrays
        return parsed_val  # Keep as is if it's not a list
    except (ValueError, SyntaxError):
        return val  # Return as is if it can't be parsed


class Replay(Node):
    @staticmethod
    def config_output_slots():
        return {"table_output": DataType.TABLE}

    @staticmethod
    def config_params():
        return {
            "Read": {
                "filename": StringParam("output.csv"),
                "play": BoolParam(False),
                "restart": BoolParam(False, trigger=True),
            }
        }

    def setup(self):
        self.df = None
        self.current_index = 0
        self.last_filename = None  # Track filename changes
        self.load_csv()

    def load_csv(self):
        import pandas as pd

        filename = self.params["Read"]["filename"].value
        if filename != self.last_filename:  # Only reload if filename changed
            self.df = pd.read_csv(
                filename,
                converters={col: convert_to_numpy for col in pd.read_csv(filename, nrows=1).columns},
            )
            self.current_index = 0
            self.last_filename = filename

    def process(self):
        # Reload CSV if filename has changed
        self.load_csv()

        if self.df is None or self.df.empty:
            return {"table_output": ({}, {})}  # Return empty table instead of None

        if not self.params["Read"]["play"].value:
            return

        # Extract the current row as a dictionary
        row_data = self.df.iloc[self.current_index].to_dict()

        # Convert each value to the appropriate Data format
        table_output = {key: to_data(value) for key, value in row_data.items()}

        # Increment index, loop back to the start when reaching the end
        self.current_index = (self.current_index + 1) % len(self.df)

        return {"table_output": (table_output, {})}

    def read_filename_changed(self):
        self.load_csv()

    def read_restart_changed(self):
        self.current_index = 0  # Reset index when restart is triggered



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import IntParam, FloatParam


class TimeDelayEmbedding(Node):
    def config_input_slots():
        return {"input_array": DataType.ARRAY}

    def config_output_slots():
        return {"embedded_array": DataType.ARRAY}

    def config_params():
        return {
            "embedding": {"delay": IntParam(1, 1, 100), "embedding_dimension": IntParam(2, 2, 100), "moire_embedding": False,
                          'exponent': FloatParam(1.0, 0.0, 10.0, doc="Exponent for time delay embedding")}
        }

    def process(self, input_array: Data):
        if input_array is None or input_array.data is None:
            return None

        delay = self.params["embedding"]["delay"].value
        embedding_dimension = self.params["embedding"]["embedding_dimension"].value
        moire_embedding = self.params["embedding"]["moire_embedding"].value

        array = input_array.data
        arrays = [array]
        exponent = self.params["embedding"]["exponent"].value
        # Generating delayed versions
        for i in range(1, embedding_dimension):
            shift = int((i ** exponent) * delay)
            delayed = np.roll(array, shift=shift, axis=0)
            arrays.append(delayed)

        if moire_embedding:
            embedded_array = np.stack(arrays, axis=-1)
        else:
            embedded_array = np.array(arrays)

        if "dim0" in input_array.meta["channels"]:
            input_array.meta["channels"]["dim1"] = input_array.meta["channels"]["dim0"]
            del input_array.meta["channels"]["dim0"]

        return {"embedded_array": (embedded_array, input_array.meta)}



import numpy as np
from numpy.fft import fft, fftfreq

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, IntParam, StringParam


class PSD(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"psd": DataType.ARRAY}

    def config_params():
        return {
            "psd": {
                "method": StringParam("welch", options=["fft", "welch"]),
                "f_min": FloatParam(-1, 0.0, 9999.0),
                "f_max": FloatParam(-1, 1.0, 10000.0),
                "axis": -1,
            },
            "welch": {"nperseg": IntParam(-1, 1, 1000), "noverlap": IntParam(-1, 0, 1000)},
        }

    def setup(self):
        from scipy.signal import welch

        self.welch = welch

    def process(self, data: Data):
        if data is None or data.data is None:
            return None

        if data.data.ndim not in [1, 2]:
            raise ValueError("Data must be 1D or 2D")

        method = self.params.psd.method.value
        f_min = self.params.psd.f_min.value
        f_max = self.params.psd.f_max.value
        axis = self.params.psd.axis.value if self.params.psd.axis.value >= 0 else data.data.ndim + self.params.psd.axis.value

        nperseg = self.params.welch.nperseg.value if self.params.welch.nperseg.value > 0 else None
        noverlap = self.params.welch.noverlap.value if self.params.welch.noverlap.value >= 0 else None

        sfreq = data.meta["sfreq"]

        if method == "fft":
            freq = fftfreq(data.data.shape[axis], 1 / sfreq)
            psd = np.abs(fft(data.data, axis=axis))
        elif method == "welch":
            freq, psd = self.welch(data.data, fs=sfreq, nperseg=nperseg, noverlap=noverlap, axis=axis)

        # selecting the range of frequencies
        if f_min < 0:
            f_min = freq.min()
        if f_max < 0:
            f_max = freq.max()
        valid_indices = np.where((freq >= f_min) & (freq <= f_max))[0]

        meta = data.meta.copy()
        freq = freq[valid_indices]
        psd = np.take(psd, valid_indices, axis=axis)
        meta["channels"][f"dim{axis}"] = freq.tolist()

        return {"psd": (psd, meta)}



from collections import deque

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, FloatParam, IntParam, StringParam


class Filter(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"filtered_data": DataType.ARRAY}

    def config_params():
        return {
            "bandpass": {
                "apply": BoolParam(False),
                "type": StringParam("butterworth", options=["butterworth", "chebyshev", "elliptic"]),
                "method": StringParam(
                    "Causal",
                    options=["Causal", "Zero-phase"],
                    doc="Zero-phase refers to applying the filter twice, once forward and once backward",
                ),
                "order": IntParam(3, 1, 10, doc="Order of the bandpass filter"),
                "f_low": FloatParam(1.0, 0.01, 9999.0),
                "f_high": FloatParam(50.0, 1.0, 10000.0),
                "ripple": FloatParam(1.0, 0.1, 10.0, doc="Ripple refers to the maximum loss in the passband of the filter"),
                "padding": FloatParam(
                    0.1,
                    0.01,
                    1.0,
                    doc="Padding refers to the fraction of the signal to pad at the beginning and end of the signal",
                ),
            },
            "notch": {
                "apply": BoolParam(False),
                "type": StringParam("butterworth", options=["butterworth", "chebyshev", "elliptic"]),
                "method": StringParam(
                    "Causal",
                    options=["Causal", "Zero-phase"],
                    doc="Zero-phase refers to applying the filter twice, once forward and once backward",
                ),
                "order": IntParam(1, 1, 4, doc="Order of the notch filter"),
                "f_center": FloatParam(60.0, 0.01, 10000.0, doc="Center frequency of the notch filter"),
                "Q": FloatParam(10.0, 0.1, 30.0, doc="Intensity of the notch filter"),
                "ripple": FloatParam(1.0, 0.1, 10.0, doc="Ripple refers to the maximum loss in the passband of the filter"),
                "padding": FloatParam(
                    0.1,
                    0.01,
                    1.0,
                    doc="Padding refers to the fraction of the signal to pad at the beginning and end of the signal",
                ),
            },
            "signal": {
                "detrend": False,
                "demean": False,
                "internal_buffer": False,
                "buffer_size": 512,
            },
        }

    def setup(self):
        from scipy import signal

        self.signal = signal

        self.filter_state_bp = None
        self.filter_state_notch = None
        self.internal_buffer = deque(maxlen=self.params["signal"]["buffer_size"].value)

    def process(self, data: Data):
        if data is None or data.data is None:
            return None

        sfreq = data.meta["sfreq"]
        nyq = 0.5 * sfreq

        self.internal_buffer.extend(data.data.T)
        if self.params["signal"]["internal_buffer"].value:
            filtered_data = np.array(self.internal_buffer).T
        else:
            filtered_data = data.data

        # Bandpass Filtering
        if self.params["bandpass"]["apply"].value:
            filter_type = self.params["bandpass"]["type"].value
            f_low = self.params["bandpass"]["f_low"].value
            f_high = self.params["bandpass"]["f_high"].value
            ripple = self.params["bandpass"]["ripple"].value
            padding = self.params["bandpass"]["padding"].value
            method = self.params["bandpass"]["method"].value
            order = self.params["bandpass"]["order"].value
            low = f_low / nyq
            high = f_high / nyq

            if filter_type == "butterworth":
                b, a = self.signal.butter(order, [low, high], btype="band")
            elif filter_type == "chebyshev":
                b, a = self.signal.cheby1(order, ripple, [low, high], btype="band")
            elif filter_type == "elliptic":
                b, a = self.signal.ellip(order, ripple, ripple, [low, high], btype="band")

            filtered_data = self.apply_filter(b, a, filtered_data, method, padding, "bandpass")

        # Notch Filtering
        if self.params["notch"]["apply"].value:
            filter_type = self.params["notch"]["type"].value
            f_center = self.params["notch"]["f_center"].value
            ripple = self.params["notch"]["ripple"].value
            padding = self.params["notch"]["padding"].value
            method = self.params["notch"]["method"].value
            order = self.params["notch"]["order"].value
            Q = self.params["notch"]["Q"].value
            w0 = f_center / nyq
            bw = f_center / Q
            bw /= nyq

            w_low = np.clip(w0 - bw / 2, 0.01, 0.99)
            w_high = np.clip(w0 + bw / 2, 0.01, 0.99)

            if filter_type == "butterworth":
                b, a = self.signal.butter(order, [w_low, w_high], btype="bandstop")
            elif filter_type == "chebyshev":
                b, a = self.signal.cheby1(order, ripple, [w_low, w_high], btype="bandstop")
            elif filter_type == "elliptic":
                b, a = self.signal.ellip(order, ripple, ripple, [w_low, w_high], btype="bandstop")

            filtered_data = self.apply_filter(b, a, filtered_data, method, padding, "notch")

        if self.params["signal"]["internal_buffer"].value:
            filtered_data = filtered_data[..., -data.data.shape[-1] :]

        # Detrend and demean
        if self.params["signal"]["detrend"].value:
            filtered_data = self.signal.detrend(filtered_data, type="linear")
        if self.params["signal"]["demean"].value:
            filtered_data = self.signal.detrend(filtered_data, type="constant")

        return {"filtered_data": (filtered_data, data.meta)}

    def apply_filter(self, b, a, data, method, padding, filter_type):
        if method == "Zero-phase":
            if data.ndim == 1:
                padlen = int(padding * len(data))
            else:
                padlen = int(padding * data.shape[-1])
            return self.signal.filtfilt(b, a, data, padlen=padlen)
        elif method == "Causal":
            if filter_type == "bandpass":
                filter_state = self.filter_state_bp
            elif filter_type == "notch":
                filter_state = self.filter_state_notch

            if data.ndim == 1:
                zi = self.signal.lfilter_zi(b, a)
                if filter_state is None or filter_state.shape != zi.shape:
                    filter_state = zi * data[0]
                filtered_data, new_state = self.signal.lfilter(b, a, data, zi=filter_state)
            elif data.ndim == 2:
                num_channels = data.shape[0]
                zi = self.signal.lfilter_zi(b, a).reshape(-1, 1)
                if filter_state is None or filter_state.shape[1] != num_channels:
                    filter_state = np.tile(zi, (1, num_channels)) * data[:, 0]
                filtered_data, new_state = self.signal.lfilter(b, a, data.T, axis=0, zi=filter_state)
                filtered_data = filtered_data.T

            # Update the appropriate filter state
            if filter_type == "bandpass":
                self.filter_state_bp = new_state
            elif filter_type == "notch":
                self.filter_state_notch = new_state

            return filtered_data
        else:
            return data

    def signal_buffer_size_changed(self, buffer_size):
        self.internal_buffer = deque(maxlen=buffer_size)



import json
from os import environ, path

import requests

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, IntParam, StringParam


class TextGeneration(Node):
    @staticmethod
    def config_input_slots():
        return {"prompt": DataType.STRING}

    @staticmethod
    def config_output_slots():
        return {"generated_text": DataType.STRING}

    @staticmethod
    def config_params():
        return {
            "text_generation": {
                "api_key": StringParam("openai.key", doc="API key for the text generation service"),
                "model": StringParam("gpt-4o-mini", doc="Model ID for the text generation service"),
                "system_prompt": StringParam("", doc="System prompt for text generation"),
                "temperature": FloatParam(1.0, 0.0, 2.0, doc="Temperature for text generation"),
                "max_tokens": IntParam(20, 5, 2048, doc="Maximum number of tokens to generate"),
                "max_conversation_length": IntParam(
                    0, 0, 100, doc="Maximum length of conversation history (0 to disable, -1 for unlimited)"
                ),
                "save_conversation": StringParam("", doc="Whether to save conversation history to a JSON file"),
            }
        }

    def setup(self):
        self.provider = None
        self.client = None
        self.messages = []
        self.text_generation_api_key_changed(self.params.text_generation.api_key.value)
        self.text_generation_model_changed(self.params.text_generation.model.value)

    def generate_openai_response(self, messages):
        if self.client is None:
            self.client = self.provider.OpenAI(api_key=self.api_key)

        payload = {
            "model": self.params.text_generation.model.value,
            "messages": messages,
            "temperature": self.params.text_generation.temperature.value,
            "max_tokens": self.params.text_generation.max_tokens.value,
        }

        if len(self.params.text_generation.system_prompt.value) > 0 and len(messages) < 2:
            payload["messages"] = [{"role": "system", "content": self.params.text_generation.system_prompt.value}] + messages

        response = self.client.chat.completions.create(**payload)
        return response.choices[0].message.content

    def generate_anthropic_response(self, messages):
        if self.client is None:
            self.client = self.provider.Anthropic(api_key=self.api_key)

        system_prompt = self.params.text_generation.system_prompt.value
        response = self.client.messages.create(
            model=self.params.text_generation.model.value,
            messages=messages,
            max_tokens=self.params.text_generation.max_tokens.value,
            temperature=self.params.text_generation.temperature.value,
            system=system_prompt if len(system_prompt) > 0 else None,
        )
        return response.content[0].text

    def generate_gemini_response(self, prompt):
        if self.client is None:
            self.provider.configure(api_key=self.api_key)
            system_prompt = self.params.text_generation.system_prompt.value
            self.client = self.provider.GenerativeModel(
                model_name=self.params.text_generation.model.value,
                system_instruction=system_prompt if len(system_prompt) > 0 else "You are a helpful assistant.",
            )

        chat = self.client.start_chat()
        generation_config = self.provider.GenerationConfig(
            max_output_tokens=self.params.text_generation.max_tokens.value,
            temperature=self.params.text_generation.temperature.value,
        )

        # send the request
        response = chat.send_message(prompt, generation_config=generation_config)
        return response.text

    def generate_ollama_response(self, model, messages):
        # create an Ollama client instance
        if self.client is None:
            self.client = self.provider

        response = self.client.chat(
            model=str(model.replace("ollama-", "")),  # remove ollama- prefix
            messages=messages,
            options={
                "system": self.params.text_generation.system_prompt.value,
                "temperature": self.params.text_generation.temperature.value,
                "max_tokens": self.params.text_generation.max_tokens.value,
            },
        )
        return response["message"]["content"]

    def generate_local_response(self, prompt):
        url = "http://127.0.0.1:5000/v1/chat/completions"
        headers = {"Content-Type": "application/json"}
        data = {
            "mode": "chat-instruct",
            "max_tokens": self.params.text_generation.max_tokens.value,
            "temperature": self.params.text_generation.temperature.value,
            "character": "Example",
            "messages": [{"role": "user", "content": prompt}],
        }
        response = requests.post(url, headers=headers, json=data, verify=False)
        return response.json()["choices"][0]["message"]["content"]

    def save_conversation_to_json(self):
        filename = f"conversation_{self.params.text_generation.save_conversation.value}.json"
        with open(filename, "w") as f:
            json.dump(self.messages, f, indent=4)
        print(f"Conversation saved to {filename}")

    def process(self, prompt: Data):
        if prompt is None or prompt.data is None:
            return None

        model = self.params.text_generation.model.value
        max_conversation_length = self.params.text_generation.max_conversation_length.value
        save_conversation = self.params.text_generation.save_conversation.value

        if max_conversation_length == 0:
            self.messages = [{"role": "user", "content": prompt.data}]
        else:
            if max_conversation_length > 0 and len(self.messages) >= max_conversation_length:
                self.messages = self.messages[-max_conversation_length + 1 :]
            self.messages.append({"role": "user", "content": prompt.data})

        if model.startswith("gpt-"):
            generated_text = self.generate_openai_response(self.messages)

        elif model.startswith("claude-"):
            generated_text = self.generate_anthropic_response(self.messages)

        elif model.startswith("gemini-"):
            if max_conversation_length != 0:
                print("The API call for Gemini currently does not support conversation history. Open an issue if needed.")
            generated_text = self.generate_gemini_response(prompt.data)

        elif model.startswith("local-"):
            if max_conversation_length != 0:
                print("The local model currently does not support conversation history. Open an issue if needed.")
            generated_text = self.generate_local_response(prompt.data)

        elif model.startswith("ollama-"):
            generated_text = self.generate_ollama_response(model, self.messages)
        else:
            raise ValueError(f"Unknown model: {model}")

        if max_conversation_length != 0:
            self.messages.append({"role": "assistant", "content": generated_text})

        if save_conversation:
            self.save_conversation_to_json()

        return {"generated_text": (generated_text, prompt.meta)}

    def text_generation_api_key_changed(self, value):
        self.client = None
        self.api_key = value

        model = self.params.text_generation.model.value
        if path.exists(self.api_key):
            with open(self.api_key, "r") as f:
                self.api_key = f.read().strip()
        elif model.startswith("gpt-"):
            if len(environ.get("OPENAI_API_KEY", "")) > 0:
                print("Using OPENAI_API_KEY from environment variables.")
            self.api_key = environ.get("OPENAI_API_KEY", self.api_key)
        elif model.startswith("claude-"):
            if len(environ.get("ANTHROPIC_API_KEY", "")) > 0:
                print("Using ANTHROPIC_API_KEY from environment variables.")
            self.api_key = environ.get("ANTHROPIC_API_KEY", self.api_key)
        elif model.startswith("gemini-"):
            if len(environ.get("GOOGLE_API_KEY", "")) > 0:
                print("Using GOOGLE_API_KEY from environment variables.")
            self.api_key = environ.get("GOOGLE_API_KEY", self.api_key)

        if not self.api_key and not model.startswith("local-"):
            raise ValueError(f"API key for {model} not found in environment variables or input parameters.")
        print(f"Loaded API key for {model}.")

    def text_generation_model_changed(self, model):
        self.provider = None
        self.client = None

        if model.startswith("gpt-"):
            try:
                import openai
            except ImportError:
                raise ImportError("You need to install openai: pip install openai")
            self.provider = openai

        elif model.startswith("claude-"):
            try:
                import anthropic
            except ImportError:
                raise ImportError("You need to install anthropic: pip install anthropic")
            self.provider = anthropic

        elif model.startswith("gemini-"):
            try:
                import google.generativeai as genai
            except ImportError:
                raise ImportError(
                    "You need to install google-generativeai: pip install google-generativeai\nWith Python>=3.10"
                )
            self.provider = genai

        elif model.startswith("ollama-"):
            try:
                import ollama
            except ImportError:
                raise ImportError("You need to install ollama: pip install ollama")
            self.provider = ollama

        else:
            raise ValueError(f"Unknown model: {model}")



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, IntParam, StringParam


class Clustering(Node):
    def config_input_slots():
        return {"matrix": DataType.ARRAY}

    def config_output_slots():
        return {
            "cluster_labels": DataType.ARRAY,
            "cluster_centers": DataType.ARRAY,  # for KMeans
        }

    def config_params():
        return {
            "Clustering": {
                "algorithm": StringParam("KMeans", options=["KMeans", "Agglomerative"]),
                "n_clusters": IntParam(2, 1, 20),  # Assuming max clusters as 20 for simplicity
                "max_iter": IntParam(300, 1, 1000, doc="Maximum number of iterations for KMeans"),
                "tolerance": FloatParam(0.0001, 0.0001, 0.1, doc="Tolerance for KMeans convergence"),
                "affinity": StringParam("euclidean", options=["euclidean", "l1", "l2", "manhattan", "cosine"]),
                "linkage": StringParam("ward", options=["ward", "complete", "average", "single"]),
            }
        }

    def setup(self):
        from sklearn.cluster import AgglomerativeClustering, KMeans

        self.AgglomerativeClustering = AgglomerativeClustering
        self.KMeans = KMeans

    def process(self, matrix: Data):
        if matrix is None:
            return None
        max_iter = self.params.Clustering["max_iter"].value
        tolerance = self.params.Clustering["tolerance"].value
        if self.params.Clustering["algorithm"].value == "KMeans":
            model = self.KMeans(n_clusters=self.params.Clustering["n_clusters"].value, max_iter=max_iter, tol=tolerance)
            labels = model.fit_predict(matrix.data)
            centers = model.cluster_centers_
            labels = labels.astype(np.float64)
            return {
                "cluster_labels": (labels, matrix.meta),
                "cluster_centers": (centers, matrix.meta),
            }

        elif self.params.Clustering["algorithm"].value == "Agglomerative":
            model = self.AgglomerativeClustering(
                n_clusters=self.params.Clustering["n_clusters"].value,
                affinity=self.params.Clustering["affinity"].value,
                linkage=self.params.Clustering["linkage"].value,
            )
            labels = model.fit_predict(matrix.data)
            # put the labels as float
            labels = labels.astype(np.float64)
            return {"cluster_labels": (labels, matrix.meta)}



from itertools import groupby

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, FloatParam, IntParam


class Recurrence(Node):
    def config_input_slots():
        return {"input_array": DataType.ARRAY}

    def config_output_slots():
        return {"recurrence_matrix": DataType.ARRAY, "RR": DataType.ARRAY, "DET": DataType.ARRAY, "LAM": DataType.ARRAY}

    def config_params():
        return {
            "recurrence": {
                "relative_threshold": FloatParam(
                    0.1,
                    0.0,
                    1.0,
                    doc="Relative threshold for binarizing the distance matrix as a fraction of the maximum distance",
                ),
                "binarize": BoolParam(True, doc="Flag to determine if distance matrix should be binarized"),
                "min_diag_length": IntParam(
                    2, 1, 100, doc="Minimum length of diagonal lines to be considered for DET and LAM calculation"
                ),
            }
        }

    def setup(self):
        from sklearn.metrics import pairwise_distances

        self.pairwise_distances = pairwise_distances

    def process(self, input_array: Data):
        if input_array is None:
            return None

        # Get the parameters
        relative_threshold = self.params["recurrence"]["relative_threshold"].value
        binarize = self.params["recurrence"]["binarize"].value

        # Check the shape of the array
        data = input_array.data
        if len(data.shape) == 1:
            # Reshape 1D array to 2D column vector
            data = data[:, np.newaxis]

        if data.shape[0] < 2:
            raise ValueError("Not enough data points to compute recurrence matrix.")

        # Compute pairwise distance
        distance_matrix = self.pairwise_distances(data)

        # Calculate the actual threshold
        threshold = relative_threshold * np.max(distance_matrix)

        # Binarize the distance matrix if required
        if binarize:
            recurrence_matrix = (distance_matrix <= threshold).astype(int)
        else:
            recurrence_matrix = distance_matrix

        # Calculate RQA metrics
        RR = np.mean(recurrence_matrix)  # Recurrence Rate
        # Get all diagonal lengths using the new method
        diagonal_lengths = self.get_diagonal_lengths(recurrence_matrix)
        DET = sum([ln**2 for ln in diagonal_lengths]) / float(recurrence_matrix.size)  # Determinism
        vertical_lengths = [
            len(list(group))
            for i in range(recurrence_matrix.shape[0])
            for key, group in groupby(recurrence_matrix[i, :])
            if key
        ]
        LAM = sum([ln**2 for ln in vertical_lengths]) / float(recurrence_matrix.size)  # Laminarity

        return {
            "recurrence_matrix": (recurrence_matrix, input_array.meta),
            "RR": (np.array(RR), {}),
            "DET": (np.array(DET), {}),
            "LAM": (np.array(LAM), {}),
        }

    def get_diagonal_lengths(self, matrix):
        N = matrix.shape[0]
        lengths = []
        for d in range(-N + 2, N - 1):  # Loop over all diagonals
            diag = np.diagonal(matrix, offset=d)
            for key, group in groupby(diag):
                if key:
                    length = len(list(group))
                    if length >= self.params["recurrence"]["min_diag_length"].value:
                        lengths.append(length)
        return lengths



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, StringParam


class RandomArray(Node):
    """
    A Goofi node that generates a random matrix based on specified dimensions and distribution.
    The dimensions are provided as a string of comma-separated integers.
    The distribution can be specified as "uniform" or "normal".
    An option is provided to set the largest eigenvalue of the matrix to 1.
    """

    def config_params():
        return {
            "random": {
                "dimensions": "2,2",  # Default dimensions
                "distribution": StringParam("normal", options=["uniform", "normal"]),
                "normalize_eigenvalue": False,  # Default: do not normalize eigenvalue
                "reset": BoolParam(trigger=True),
            },
            "common": {
                "autotrigger": True,
            },
        }

    def config_output_slots():
        return {"random_array": DataType.ARRAY}

    def setup(self):
        self.random_array = None

        distribution = self.params.random.distribution.value
        normalize_eigenvalue = self.params.random.normalize_eigenvalue.value

        try:
            # Parse the dimensions string into a tuple of integers
            dimensions = tuple(map(int, self.params.random.dimensions.value.split(",")))

            # Generate a random array based on the specified distribution
            if distribution == "uniform":
                random_array = np.random.uniform(size=dimensions)
            elif distribution == "normal":
                random_array = np.random.normal(0, 1, size=dimensions)
            else:
                raise ValueError(f"Unsupported distribution: {distribution}. Use 'uniform' or 'normal'.")

            # Normalize the largest eigenvalue to 1 if specified
            if normalize_eigenvalue and len(dimensions) == 2 and dimensions[0] == dimensions[1]:
                eigenvalues, _ = np.linalg.eig(random_array)
                max_eigenvalue = np.max(np.abs(eigenvalues))
                if max_eigenvalue != 0:
                    random_array /= max_eigenvalue

            self.random_array = random_array

        except (ValueError, TypeError) as e:
            # Handle errors in parsing dimensions or generating the array
            raise ValueError("Invalid dimensions format or distribution type. Please provide a valid input.") from e

    def process(self):
        return {"random_array": (self.random_array, {})}

    def random_dimensions_changed(self, val):
        self.setup()

    def random_distribution_changed(self, val):
        self.setup()

    def random_normalize_eigenvalue_changed(self, val):
        self.setup()

    def random_reset_changed(self, val):
        self.setup()



import time

import numpy as np

from goofi.data import DataType
from goofi.node import Node
from goofi.params import FloatParam


class Sine(Node):
    def config_params():
        return {
            "sine": {"frequency": FloatParam(1.0, 0.1, 30.0), "sampling_frequency": FloatParam(1000.0, 1.0, 1000.0)},
            "common": {"autotrigger": True},
        }

    def config_input_slots():
        return {"frequency": DataType.ARRAY}

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def setup(self):
        self.last_trigger = time.time()

    def process(self, frequency):
        freq = frequency.data.item() if frequency else self.params.sine.frequency.value

        meta = {"sfreq": self.params.sine.sampling_frequency.value}

        t = time.time()
        dt = t - self.last_trigger
        xs = np.arange(t, t + dt, 1 / self.params.sine.sampling_frequency.value)
        data = np.sin(xs * np.pi * 2 * freq)

        self.last_trigger = t
        return {"out": (data, meta)}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import StringParam


class LempelZiv(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"lzc": DataType.ARRAY}

    def config_params():
        return {
            "lempel_ziv": {
                "binarization": StringParam("mean", options=["mean", "median"]),
                "axis": -1,
            }
        }

    def setup(self):
        from antropy import lziv_complexity

        self.compute_lzc = lziv_complexity

    def process(self, data: Data):
        if data is None:
            # no data, skip processing
            return None

        # read parameters
        binarize_mode = self.params.lempel_ziv.binarization.value
        axis = self.params.lempel_ziv.axis.value

        # binarize data
        if binarize_mode == "mean":
            binarized = data.data > np.mean(data.data, axis=axis, keepdims=True)  # mean split
        elif binarize_mode == "median":
            binarized = data.data > np.median(data.data, axis=axis, keepdims=True)  # median split

        # compute Lempel-Ziv complexity
        lzc = np.apply_along_axis(self.compute_lzc, axis, binarized, normalize=True)
        #                         ^^^^^^^^^^^^^^^^  ^^^^   ^^^^^^^^^  ^^^^^^^^^^^^^^
        #                         fn to apply       axis   data       args to fn

        # return Lempel-Ziv complexity and incoming metadata
        return {"lzc": (lzc, data.meta)}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, IntParam, StringParam


class TransitionalHarmony(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"trans_harm": DataType.ARRAY, "melody": DataType.ARRAY}

    def config_params():
        return {
            "transitional_harmony": {
                "n_peaks": IntParam(5, 1, 10),
                "f_min": FloatParam(2.0, 0.1, 50.0),
                "f_max": FloatParam(30.0, 1.0, 100.0),
                "precision": FloatParam(0.1, 0.01, 10.0),
                "peaks_function": StringParam("EMD", options=["EMD", "fixed", "harmonic_recurrence", "EIMC", "cepstrum"]),
                "delta": IntParam(50, 1, 250),
                "subharmonics": IntParam(10, 2, 100),
            }
        }

    def setup(self):
        from biotuner.biotuner_object import compute_biotuner
        from biotuner.metrics import compute_subharmonics_2lists

        self.compute_biotuner = compute_biotuner
        self.compute_subharmonics_2lists = compute_subharmonics_2lists

    def process(self, data: Data):
        if data is None:
            return None

        data.data = np.squeeze(data.data)
        if data.data.ndim > 1:
            raise ValueError("Data must be 1D")

        # split data in two
        data_len = len(data.data)
        data1 = data.data[: int(data_len / 2)]
        data2 = data.data[int(data_len / 2) :]

        # extract peaks from data1
        bt1 = self.compute_biotuner(
            sf=data.meta["sfreq"],
            peaks_function=self.params.transitional_harmony.peaks_function.value,
        )
        bt1.peaks_extraction(
            data1,
            n_peaks=self.params.transitional_harmony.n_peaks.value,
            min_freq=self.params.transitional_harmony.f_min.value,
            max_freq=self.params.transitional_harmony.f_max.value,
            precision=self.params.transitional_harmony.precision.value,
        )

        # extract peaks from data2
        bt2 = self.compute_biotuner(
            sf=data.meta["sfreq"],
            peaks_function=self.params.transitional_harmony.peaks_function.value,
        )
        bt2.peaks_extraction(
            data2,
            n_peaks=self.params.transitional_harmony.n_peaks.value,
            min_freq=self.params.transitional_harmony.f_min.value,
            max_freq=self.params.transitional_harmony.f_max.value,
            precision=self.params.transitional_harmony.precision.value,
        )

        # compute subharmonics between peaks
        result = self.compute_subharmonics_2lists(
            bt1.peaks,
            bt2.peaks,
            self.params.transitional_harmony.subharmonics.value,
            delta_lim=self.params.transitional_harmony.delta.value,
            c=2.1,
        )

        common_subs, delta_t, sub_tension_final, harm_temp, pairs_melody = result

        return {
            "trans_harm": (np.array(sub_tension_final), data.meta),
            "melody": (np.array(pairs_melody), data.meta),
        }



import tempfile

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam


class Monolith(Node):

    def config_params():
        return {
            "monolith": {
                "toto": BoolParam(False, doc="Whether to include Toto embedings"),
                "pyspi": BoolParam(False, doc="Whether to compute SPI features"),
            }
        }

    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"features": DataType.ARRAY}

    def setup(self):
        pass

    def process(self, data: Data):
        assert "sfreq" in data.meta, "Data must have a 'sfreq' (sampling frequency) in its metadata."
        sfreq = data.meta["sfreq"]
        data_arr = data.data

        feat_fns = FEAT_FNS.copy()
        if not self.params.monolith.pyspi.value:
            feat_fns.pop("spi", None)
        if not self.params.monolith.toto.value:
            feat_fns.pop("toto", None)

        # Preprocess the data
        data_arr = preprocess(data_arr, sfreq)

        # compute channel-wise features
        features = []
        for feat_name, feat_fn in feat_fns.items():
            if feat_name in NON_CHANNEL_WISE_FEATS:
                ft = feat_fn(data_arr, sfreq)
                if isinstance(ft, np.ndarray) or isinstance(ft, list):
                    features.append(ft)
                else:
                    features.append([ft])
            else:
                for channel_data in data_arr:
                    ft = feat_fn(channel_data, sfreq)
                    if isinstance(ft, np.ndarray) or isinstance(ft, list):
                        features.append(ft)
                    else:
                        features.append([ft])

        features = np.concatenate(features, axis=-1)
        meta = data.meta.copy()
        del meta["channels"]
        return {"features": (features, meta)}


def preprocess(data: np.ndarray, sfreq: float):
    from scipy.signal import butter, filtfilt, iirnotch

    # Apply bandpass filter (3-30 Hz)
    nyquist = sfreq / 2
    low_cut = 3 / nyquist
    high_cut = 30 / nyquist
    b, a = butter(4, [low_cut, high_cut], btype="band")
    data = filtfilt(b, a, data, axis=-1)

    # Apply notch filters for 50Hz and harmonics
    for freq in np.arange(50, 101, 50):
        if freq < nyquist:
            b, a = iirnotch(freq, Q=30, fs=sfreq)
            data = filtfilt(b, a, data, axis=-1)

    # Clip values
    data = np.clip(data, -150, 150)

    # Standardize per channel
    data = (data - np.mean(data, axis=-1, keepdims=True)) / np.std(data, axis=-1, keepdims=True)

    return data


def mean_amplitude(x, sfreq):
    return np.mean(x)


def std_amplitude(x, sfreq):
    return np.std(x)


def skewness(x, sfreq):
    from scipy.stats import skew

    return skew(x)


def kurt(x, sfreq):
    from scipy.stats import kurtosis

    return kurtosis(x)


def zero_crossings(x, sfreq):
    return ((x[:-1] * x[1:]) < 0).sum()


def hjorth_activity(x, sfreq):
    return np.var(x)


def hjorth_mobility(x, sfreq):
    return np.sqrt(np.var(np.diff(x)) / np.var(x))


def hjorth_complexity(x, sfreq):
    diff1 = np.diff(x)
    diff2 = np.diff(diff1)
    return (np.sqrt(np.var(diff2) / np.var(diff1))) / (np.sqrt(np.var(diff1) / np.var(x)))


def spectral(x, sfreq):
    from scipy.signal import welch

    f, Pxx = welch(x, sfreq)
    mean_frequency = np.sum(f * Pxx) / np.sum(Pxx)
    delta = np.trapezoid(Pxx[(f >= 0) & (f <= 4)], f[(f >= 0) & (f <= 4)])
    theta = np.trapezoid(Pxx[(f > 4) & (f <= 8)], f[(f > 4) & (f <= 8)])
    alpha = np.trapezoid(Pxx[(f > 8) & (f <= 12)], f[(f > 8) & (f <= 12)])
    beta = np.trapezoid(Pxx[(f > 12) & (f <= 30)], f[(f > 12) & (f <= 30)])
    gamma = np.trapezoid(Pxx[(f > 30) & (f <= 45)], f[(f > 30) & (f <= 45)])

    from fooof import FOOOF

    model = FOOOF(peak_width_limits=(2, 12))
    model.fit(f, Pxx, freq_range=(3, 30))

    return np.array([delta, theta, alpha, beta, gamma, mean_frequency] + list(model.aperiodic_params_))


def compute_detrended_fluctuation(x, sfreq):
    from antropy import detrended_fluctuation

    return detrended_fluctuation(x)


def compute_higuchi_fd(x, sfreq):
    from antropy import higuchi_fd

    return higuchi_fd(x)


def compute_lziv_complexity(x, sfreq):
    from antropy import lziv_complexity

    return lziv_complexity(x > x.mean(), normalize=True)


def compute_petrosian_fd(x, sfreq):
    from antropy import petrosian_fd

    return petrosian_fd(x)


def binarize_by_mean(x):
    return (x > np.mean(x)).astype(int)


def entropy_shannon(x, sfreq):
    import neurokit2 as nk2

    bin_ts = binarize_by_mean(x)
    return nk2.entropy_shannon(bin_ts, base=2)[0]


def entropy_renyi(x, sfreq):
    import neurokit2 as nk2

    bin_ts = binarize_by_mean(x)
    return nk2.entropy_renyi(bin_ts, alpha=2)[0]


def entropy_approximate(x, sfreq):
    import neurokit2 as nk2

    return nk2.entropy_approximate(x, delay=1, dimension=2, tolerance="sd", Corrected=True)[0]


def entropy_sample(x, sfreq):
    import neurokit2 as nk2

    return nk2.entropy_sample(x, delay=1, dimension=2, tolerance="sd")[0]


def entropy_rate(x, sfreq):
    import neurokit2 as nk2

    return nk2.entropy_rate(x, kmax=10, symbolize="mean")[0]


def entropy_permutation(x, sfreq):
    import neurokit2 as nk2

    return nk2.entropy_permutation(x, delay=1, dimension=2, corrected=True, weighted=False, conditional=False)[0]


def entropy_permutation_weighted(x, sfreq):
    import neurokit2 as nk2

    return nk2.entropy_permutation(x, delay=1, dimension=2, corrected=True, weighted=True, conditional=False)[0]


def entropy_permutation_conditional(x, sfreq):
    import neurokit2 as nk2

    return nk2.entropy_permutation(x, delay=1, dimension=2, corrected=True, weighted=False, conditional=True)[0]


def entropy_permutation_weighted_conditional(x, sfreq):
    import neurokit2 as nk2

    return nk2.entropy_permutation(x, delay=1, dimension=2, corrected=True, weighted=True, conditional=True)[0]


def entropy_multiscale(x, sfreq):
    import neurokit2 as nk2

    return nk2.entropy_multiscale(x, dimension=2, tolerance="sd", method="MSPEn")[0]


def entropy_bubble(x, sfreq):
    import neurokit2 as nk2

    return nk2.entropy_bubble(x, delay=1, dimension=2, alpha=2, tolerance="sd")[0]


def entropy_svd(x, sfreq):
    import neurokit2 as nk2

    return nk2.entropy_svd(x, delay=1, dimension=2)[0]


def entropy_attention(x, sfreq):
    import neurokit2 as nk2

    return nk2.entropy_attention(x)[0]


def entropy_dispersion(x, sfreq):
    import neurokit2 as nk2

    return nk2.entropy_dispersion(x, delay=1, dimension=2, c=6, symbolize="NCDF")[0]


def compute_spi_features(data, subset="fast"):
    from pyspi.calculator import Calculator

    # Create a temporary file with the config
    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
        f.write(PYSPI_FAST_CONFIG)
        pyspi_fast_config = f.name

    calc = Calculator(data, subset=subset, configfile=pyspi_fast_config)
    calc.compute()
    s = calc.table.stack(list(range(calc.table.columns.nlevels))).dropna()
    print(s, s.values)
    return s.values


def phiid_pairwise_metrics(x, sfreq):
    """
    Compute all PhiID atom metrics for all pairs of channels (asymmetric, src->tgt and tgt->src).
    Returns a flattened array of all atom means for each channel pair and direction.
    """
    from phyid.calculate import calc_PhiID

    data = np.atleast_2d(x)
    n_channels, n_time = data.shape
    tau = 5
    kind = "gaussian"
    redundancy = "MMI"
    atom_names = [
        "rtr",
        "rtx",
        "rty",
        "rts",
        "xtr",
        "xtx",
        "xty",
        "xts",
        "ytr",
        "ytx",
        "yty",
        "yts",
        "str",
        "stx",
        "sty",
        "sts",
    ]
    # For each ordered pair (i, j), i != j
    results = []
    for i in range(n_channels):
        for j in range(n_channels):
            if i == j:
                continue
            src = data[i]
            trg = data[j]
            atoms_res, _ = calc_PhiID(src, trg, tau, kind=kind, redundancy=redundancy)
            atom_means = [float(np.mean(atoms_res[name])) for name in atom_names]
            results.append(atom_means)
    return np.array(results, dtype=np.float32).flatten()


def compute_toto_embedding(x, sfreq):
    from toto.inference.embedding import embed as embed_toto

    return embed_toto(x, global_average=True)


FEAT_FNS = {
    "detrended_fluctuation": compute_detrended_fluctuation,
    "higuchi_fd": compute_higuchi_fd,
    "lziv_complexity": compute_lziv_complexity,
    "petrosian_fd": compute_petrosian_fd,
    "mean_amplitude": mean_amplitude,
    "std_amplitude": std_amplitude,
    "skewness": skewness,
    "kurtosis": kurt,
    "zero_crossings": zero_crossings,
    "hjorth_activity": hjorth_activity,
    "hjorth_mobility": hjorth_mobility,
    "hjorth_complexity": hjorth_complexity,
    "spectral": spectral,
    "entropy_shannon": entropy_shannon,
    "entropy_renyi": entropy_renyi,
    # "entropy_approximate": entropy_approximate,
    # "entropy_sample": entropy_sample,
    "entropy_rate": entropy_rate,
    # "entropy_permutation": entropy_permutation,
    # "entropy_permutation_weighted": entropy_permutation_weighted,
    # "entropy_permutation_conditional": entropy_permutation_conditional,
    # "entropy_permutation_weighted_conditional": entropy_permutation_weighted_conditional,
    # "entropy_multiscale": entropy_multiscale,
    "entropy_bubble": entropy_bubble,
    "entropy_svd": entropy_svd,
    "entropy_attention": entropy_attention,
    "entropy_dispersion": entropy_dispersion,
    "spi": compute_spi_features,
    "phiid": phiid_pairwise_metrics,
    "toto": compute_toto_embedding,
}

BATCHED_FEATS = ["toto"]
NON_CHANNEL_WISE_FEATS = ["spi", "phiid", "toto"]

PYSPI_FAST_CONFIG = """
.statistics.basic:
  Covariance:
    labels:
    - undirected
    - linear
    - signed
    - multivariate
    - contemporaneous
    dependencies: null
    configs:
    - estimator: ShrunkCovariance
      squared: true
  Precision:
    labels:
    - undirected
    - linear
    - signed
    - multivariate
    - contemporaneous
    dependencies: null
    configs:
    - estimator: ShrunkCovariance
      squared: true
  SpearmanR:
    labels:
    - undirected
    - nonlinear
    - signed
    - bivariate
    - contemporaneous
    dependencies: null
    configs:
    - squared: false
  KendallTau:
    labels:
    - undirected
    - nonlinear
    - signed
    - bivariate
    - contemporaneous
    dependencies: null
    configs:
    - squared: false
  CrossCorrelation:
    labels:
    - undirected
    - linear
    - signed/unsigned
    - bivariate
    - time-dependent
    dependencies: null
    configs:
    - statistic: mean
      squared: true
      sigonly: false
.statistics.distance:
  PairwiseDistance:
    labels:
    - unsigned
    - unordered
    - nonlinear
    - undirected
    dependencies: null
    configs:
    - metric: braycurtis
  DynamicTimeWarping:
    labels:
    - undirected
    - nonlinear
    - unsigned
    - bivariate
    - time-dependent
    dependencies: null
    configs:
    - global_constraint: sakoe_chiba
  LongestCommonSubsequence:
    labels:
    - undirected
    - nonlinear
    - unsigned
    - bivariate
    - time-dependent
    dependencies: null
    configs:
    - global_constraint: sakoe_chiba
.statistics.infotheory:
  JointEntropy:
    labels:
    - undirected
    - nonlinear
    - unsigned
    - bivariate
    - contemporaneous
    dependencies:
    - java
    configs:
    - estimator: kernel
  ConditionalEntropy:
    labels:
    - undirected
    - nonlinear
    - unsigned
    - bivariate
    - contemporaneous
    dependencies:
    - java
    configs:
    - estimator: kernel
  CrossmapEntropy:
    labels:
    - unsigned
    - directed
    - time-dependent
    - bivariate
    dependencies:
    - java
    configs:
    - estimator: gaussian
      history_length: 10
  StochasticInteraction:
    labels:
    - undirected
    - nonlinear
    - unsigned
    - bivariate
    - time-dependent
    dependencies:
    - java
    configs:
    - estimator: gaussian
  MutualInfo:
    labels:
    - undirected
    - nonlinear
    - unsigned
    - bivariate
    - contemporaneous
    dependencies:
    - java
    configs:
    - estimator: gaussian
  TimeLaggedMutualInfo:
    labels:
    - directed
    - nonlinear
    - unsigned
    - bivariate
    - time-dependent
    dependencies:
    - java
    configs:
    - estimator: gaussian
.statistics.spectral:
  CoherencePhase:
    labels:
    - undirected
    - linear
    - unsigned
    - bivariate
    - frequency-dependent
    dependencies: null
    configs:
    - fmin: 0.25
      fmax: 0.5
      statistic: max
  CoherenceMagnitude:
    labels:
    - undirected
    - linear
    - unsigned
    - bivariate
    - frequency-dependent
    dependencies: null
    configs:
    - fmin: 0.25
      fmax: 0.5
      statistic: max
  ImaginaryCoherence:
    labels:
    - undirected
    - linear
    - unsigned
    - bivariate
    - frequency-dependent
    dependencies: null
    configs:
    - fmin: 0.25
      fmax: 0.5
      statistic: max
  PhaseSlopeIndex:
    labels:
    - directed
    - linear/nonlinear
    - unsigned
    - bivariate
    - frequency-dependent
    dependencies: null
    configs:
    - fmin: 0.25
      fmax: 0.5
  PhaseLockingValue:
    labels:
    - undirected
    - linear
    - unsigned
    - bivariate
    - frequency-dependent
    dependencies: null
    configs:
    - fmin: 0.25
      fmax: 0.5
      statistic: max
  PhaseLagIndex:
    labels:
    - undirected
    - linear
    - unsigned
    - bivariate
    - frequency-dependent
    dependencies: null
    configs:
    - fmin: 0.25
      fmax: 0.5
      statistic: max
  WeightedPhaseLagIndex:
    labels:
    - undirected
    - linear
    - unsigned
    - bivariate
    - frequency-dependent
    dependencies: null
    configs:
    - fmin: 0.25
      fmax: 0.5
      statistic: max
  DebiasedSquaredPhaseLagIndex:
    labels:
    - undirected
    - linear
    - unsigned
    - bivariate
    - frequency-dependent
    dependencies: null
    configs:
    - fmin: 0.25
      fmax: 0.5
      statistic: max
  DebiasedSquaredWeightedPhaseLagIndex:
    labels:
    - undirected
    - linear
    - unsigned
    - bivariate
    - frequency-dependent
    dependencies: null
    configs:
    - fmin: 0.25
      fmax: 0.5
      statistic: max
  PairwisePhaseConsistency:
    labels:
    - undirected
    - linear
    - unsigned
    - bivariate
    - frequency-dependent
    dependencies: null
    configs:
    - fmin: 0.25
      fmax: 0.5
      statistic: max
  GroupDelay:
    labels:
    - directed
    - linear
    - unsigned
    - bivariate
    - frequency-dependent
    dependencies: null
    configs:
    - fmin: 0.25
      fmax: 0.5
      statistic: delay
  SpectralGrangerCausality:
    labels:
    - directed
    - linear
    - unsigned
    - bivariate
    - frequency-dependent
    dependencies: null
    configs:
    - fmin: 0.25
      fmax: 0.5
      order: 20
      method: parametric
      statistic: max
"""



import numpy as np
from numpy.fft import ifft
from goofi.data import Data, DataType
from goofi.node import Node


class IFFT(Node):
    def config_input_slots():
        return {"spectrum": DataType.ARRAY, "phase": DataType.ARRAY}

    def config_output_slots():
        return {"reconstructed": DataType.ARRAY}

    def config_params():
        return {}  # No parameters for this node

    def process(self, spectrum: Data, phase: Data):
        # Check if the input is provided
        if spectrum is None or spectrum.data is None or phase is None or phase.data is None:
            return None

        # Check if the lengths of spectrum and phase are different
        if len(spectrum.data) != len(phase.data):
            # Calculate the difference in lengths
            length_diff = abs(len(spectrum.data) - len(phase.data))

            # Check which data is shorter and pad accordingly
            if len(spectrum.data) < len(phase.data):
                # Pad the spectrum data with zeros (or another value) to match the phase data length
                padding = np.zeros(length_diff)
                spectrum.data = np.concatenate([spectrum.data, padding])
            else:
                # Pad the phase data with zeros (or another value) to match the spectrum data length
                padding = np.zeros(length_diff)
                phase.data = np.concatenate([phase.data, padding])

        # Create complex numbers from magnitude (spectrum) and phase
        complex_data = spectrum.data * np.exp(1j * phase.data)

        # Inverse Fourier Transform to get the time series
        time_series = np.real(ifft(complex_data))

        # For this example, I'm not adjusting or copying metadata, but you can do so as needed
        return {"reconstructed": (time_series, phase.meta)}



from typing import Any, Dict, Tuple

from goofi.node import Node
from goofi.params import FloatParam


class EEGRecording(Node):

    def config_params():
        return {
            "recording": {
                "use_example_data": True,
                "file_path": "",
                "file_sfreq": FloatParam(256, vmax=1000),
                "source_name": "goofi",
                "stream_name": "recording",
            }
        }

    def setup(self):
        """
        Load the data and start the stream.
        """
        import mne
        import pandas as pd
        from mne.datasets import eegbci
        from mne_lsl.player import PlayerLSL

        # stop previous stream if it exists
        self.stop()

        if self.params.recording.use_example_data.value:
            if self.params.recording.file_path.value != "":
                # both use_example_data and file_path are set
                # TODO: add proper logging
                print(
                    f"Both 'use_example_data' and 'file_path' are set. Prioritizing file: {self.params.recording.file_path.value}"
                )

        assert self.params.recording.stream_name.value != "", "Stream name cannot be empty."

        if self.params.recording.file_path.value != "":
            if self.params.recording.file_path.value.endswith(".csv"):
                # load data from csv file
                df = pd.read_csv(self.params.recording.file_path.value, index_col=0)
                df = df.select_dtypes(include=["float"])
                data = df.transpose().to_numpy()

                sfreq = self.params.recording.file_sfreq.value
                sfreq = sfreq if sfreq > 0 else 256
                info = mne.create_info(ch_names=df.columns.tolist(), ch_types=["eeg"] * data.shape[0], sfreq=sfreq)
                raw = mne.io.RawArray(data, info)
            else:
                # load data from an MNE-compatible file
                raw = mne.io.read_raw(self.params.recording.file_path.value, preload=True)
        elif self.params.recording.use_example_data.value:
            raw = mne.concatenate_raws(
                [mne.io.read_raw(p, preload=True, verbose=False) for p in eegbci.load_data(1, [1, 2], update_path=False)],
                verbose=False,
            )
            eegbci.standardize(raw)
            # scale the data for better default behavior
            raw.apply_function(lambda x: x * 1e4)
        else:
            raise RuntimeError("No data source specified. Set either 'use_example_data' or 'file_path'.")

        # start the stream
        self.stream = PlayerLSL(
            raw,
            name=self.params.recording.stream_name.value,
            source_id=self.params.recording.source_name.value,
            annotations=False,
        )
        self.stream.start()

        # NOTE: this is a special case since the node doesn't process data, so errors are never cleared
        self.clear_error()

    def stop(self):
        """Stop the stream if it exists."""
        if hasattr(self, "stream") and self.stream is not None:
            self.stream.stop()
            self.stream = None

    def process(self) -> Dict[str, Tuple[Any, Dict[str, Any]]]:
        """Should never run, as the stream runs on its own."""
        raise NotImplementedError

    def recording_use_example_data_changed(self, _):
        """Reinitialize the stream."""
        self.setup()

    def recording_file_path_changed(self, _):
        """Reinitialize the stream."""
        self.setup()

    def recording_file_sfreq_changed(self, _):
        """Reinitialize the stream."""
        self.setup()

    def recording_source_name_changed(self, _):
        """Reinitialize the stream."""
        self.setup()

    def recording_stream_name_changed(self, _):
        """Reinitialize the stream."""
        self.setup()

    def terminate(self):
        """Stop the stream and terminate the node."""
        self.stop()



from multiprocessing import shared_memory

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node


class SharedMemOut(Node):
    def config_params():
        return {"shared_memory": {"name": "goofi-pipe-memory"}}

    def config_input_slots():
        return {"data": DataType.ARRAY}

    def setup(self):
        self.shm = None

    def process(self, data: Data):
        if data is None:
            return

        # float32 = 4 bytes
        size = np.prod(data.data.shape) * 4

        if self.shm is None or self.shm.size != size or self.shm.buf is None:
            # create the shared memory
            self.create_shared_memory(size)

        if self.shm is None:
            raise RuntimeError("Shared memory could not be created.")

        # copy the data to the shared memory
        data = data.data.astype(np.float32)
        self.shm.buf[:] = data.tobytes()

    def create_shared_memory(self, size: int):
        if self.shm is not None:
            # close and unlink the existing shared memory
            self.shm.close()
            self.shm.unlink()

        name = self.params.shared_memory.name.value

        try:
            # try to create the shared memory
            self.shm = shared_memory.SharedMemory(name, create=True, size=int(size))
        except FileExistsError:
            # try to close and unlink the existing shared memory
            self.shm = shared_memory.SharedMemory(name, create=False)
            self.shm.close()
            self.shm.unlink()
            # try to create the shared memory again
            self.shm = shared_memory.SharedMemory(name, create=True, size=int(size))

    def shared_memory_name_changed(self, value):
        if self.shm is not None:
            self.shm.close()
            self.shm.unlink()



import numpy as np
from goofi.data import DataType
from goofi.node import Node
from goofi.params import FloatParam

class FrequencyShift(Node):
    @staticmethod
    def config_input_slots():
        return {"data": DataType.ARRAY}

    @staticmethod
    def config_output_slots():
        return {"out": DataType.ARRAY}

    @staticmethod
    def config_params():
        return {
            "shift": {
                "frequency_shift": FloatParam(
                    1.0, -1000.0, 1000.0, doc="Frequency shift in Hz"
                ),
            }
        }

    def process(self, data):
        # early exit
        if data is None or data.data is None:
            return None

        # flatten to 1D
        signal = np.asarray(data.data).flatten()
        sfreq = data.meta.get("sfreq")
        if sfreq is None:
            raise ValueError("data.meta must include sampling frequency 'sfreq'.")

        n = signal.shape[0]
        freq_shift = self.params["shift"]["frequency_shift"].value

        # compute how many FFT bins to shift
        delta_bins = int(round(freq_shift * n / sfreq))

        # perform FFT
        spectrum = np.fft.fft(signal)

        # allocate shifted spectrum
        shifted_spectrum = np.zeros_like(spectrum)

        # positive shift => move content upward
        if delta_bins > 0:
            shifted_spectrum[delta_bins:] = spectrum[:-delta_bins]
        # negative shift => move content downward
        elif delta_bins < 0:
            shifted_spectrum[:delta_bins] = spectrum[-delta_bins:]
        else:
            shifted_spectrum[:] = spectrum

        # invert back to time domain
        shifted_signal = np.fft.ifft(shifted_spectrum)
        # because spectrum shift breaks conjugate symmetry, take real part
        shifted_signal = np.real(shifted_signal)

        # trim/pad to original length (just in case)
        shifted_signal = shifted_signal[:n]

        return {"out": (shifted_signal, data.meta)}



from copy import deepcopy

import numpy as np
from numpy.fft import fft, fftfreq

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import StringParam


class FFT(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"mag": DataType.ARRAY, "phase": DataType.ARRAY}

    def config_params():
        return {
            "fft": {
                "input_type": StringParam("time_series", options=["time_series", "image"]),
            }
        }

    def process(self, data: Data):
        if data is None or data.data is None:
            return None

        input_type = self.params["fft"]["input_type"].value

        if input_type == "time_series":
            if data.data.ndim not in [1, 2]:
                raise ValueError("Data must be 1D or 2D")
            sfreq = data.meta["sfreq"]
            freq = fftfreq(data.data.shape[-1], 1 / sfreq)
            fft_result = fft(data.data, axis=-1)
            psd = np.abs(fft_result)
            phase = np.angle(fft_result)

            meta = deepcopy(data.meta)
            if data.data.ndim == 1:
                meta["channels"]["dim0"] = freq.tolist()
            else:  # if 2D
                meta["channels"]["dim1"] = freq.tolist()

        elif input_type == "image":
            if data.data.ndim == 2:  # Grayscale image
                fft_result = np.fft.fft2(data.data)
            elif data.data.ndim == 3 and data.data.shape[2] == 3:  # RGB image
                fft_result = np.stack([np.fft.fft2(data.data[:, :, channel]) for channel in range(3)], axis=-1)
            else:
                raise ValueError("Unsupported image shape for FFT. Supported shapes are (x, y) or (x, y, 3)")

            psd = np.abs(fft_result)
            phase = np.angle(fft_result)

            # Meta data remains the same for image since we're not really modifying the "channels" in the image case
            meta = deepcopy(data.meta)

        return {"mag": (psd, meta), "phase": (phase, meta)}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import StringParam


class Function(Node):
    def config_input_slots():
        return {"array": DataType.ARRAY}

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def config_params():
        return {
            "function": {
                "function": StringParam(
                    "abs",
                    options=["abs", "sqrt", "log", "exp", "sin", "cos", "tan", "arcsin", "arccos", "arctan"],
                    doc="Element-wise function to apply to the array data.",
                )
            }
        }

    def process(self, array: Data):
        if array is None:
            return None

        func = getattr(np, self.params.function.function.value, None)
        if func is None:
            raise ValueError(f"Unknown function: {self.params.function.function.value}")
        result = func(array.data)

        return {"out": (result, array.meta)}



import numpy as np
from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, StringParam, IntParam


class Kuramoto(Node):
    def config_input_slots():
        return {"initial_phases": DataType.ARRAY}

    def config_output_slots():
        return {
            "phases": DataType.ARRAY,
            "coupling": DataType.ARRAY,
            "order_parameter": DataType.ARRAY,
            "waveforms": DataType.ARRAY,
        }

    def config_params():
        return {
            "kuramoto": {
                "coupling_strength": FloatParam(0.5, 0.0, 10.0),
                "natural_frequencies": StringParam(
                    "1, 1.5, 2", doc="Comma-separated list of natural frequencies"
                ),
                "timesteps": IntParam(
                    1000, 0, 10000, doc="Number of timesteps to integrate"
                ),
            }
        }

    def process(self, initial_phases: Data):
        # Parse the string to get the natural frequencies
        omega_str = self.params["kuramoto"]["natural_frequencies"].value
        omega = np.array([float(freq.strip()) for freq in omega_str.split(",")])

        N = len(omega)  # Determine the number of oscillators from the length of omega
        K = self.params["kuramoto"]["coupling_strength"].value
        dt = 0.01

        # Number of timesteps to integrate
        timesteps = self.params["kuramoto"]["timesteps"].value

        if initial_phases is None:
            theta = 2 * np.pi * np.random.rand(N)
        else:
            theta = initial_phases.data

        # Store phases at each timestep
        theta_history = np.zeros((N, timesteps))
        time_points = np.linspace(0, dt * timesteps, timesteps)

        # Integrate over the specified number of timesteps
        for i in range(timesteps):
            coupling_term = np.sum(np.sin(theta - theta[:, np.newaxis]), axis=1) / N
            dtheta = dt * (omega + K * coupling_term)
            theta += dtheta
            theta = np.mod(theta, 2 * np.pi)
            theta_history[:, i] = theta

        # Compute actual waveforms from the phase history
        waveforms = np.sin(np.outer(omega, time_points) + theta_history)

        # Calculate the order parameter
        R = np.abs(np.mean(np.exp(1j * theta)))

        channels = {"dim0": [f"oscillator_{i}" for i in range(N)]}
        meta = {"channels": channels}
        # add sfreq to meta
        sfreq = 1 / dt
        meta["sfreq"] = sfreq

        return {
            "phases": (theta.reshape(-1, 1), meta),
            "coupling": (coupling_term.reshape(-1, 1), meta),
            "order_parameter": (np.array(R), {}),
            "waveforms": (waveforms, meta),
        }



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam


class Avalanches(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {
            "size": DataType.ARRAY,
            "duration": DataType.ARRAY,
        }

    def config_params():
        return {
            "parameters": {
                "time_bin": FloatParam(0.008, 0.0, 0.05),
            }
        }

    def setup(self):
        import edgeofpy as eop

        self.eop = eop

    def process(self, data: Data):
        if data.data is None:
            return None

        data.data = np.squeeze(data.data)
        if data.data.ndim == 1:
            # create a new axis if the data is 1D
            data.data = data.data[np.newaxis, :]
        time_bin = self.params["parameters"]["time_bin"].value
        avalanches, _, _ = self.eop.avalanche._det_avls(data.data, s_freq=data.meta["sfreq"], time=None, max_iei=time_bin)
        sizes = np.array([np.float(x["size"]) for x in avalanches])
        durations = np.array([x["dur_sec"] for x in avalanches])
        return {"size": (sizes, {}), "duration": (durations, {})}



from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import StringParam


class Reshape(Node):
    def config_input_slots():
        return {"array": DataType.ARRAY}

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def config_params():
        return {"reshape": {"shape": StringParam("-1")}}

    def process(self, array: Data):
        if array is None:
            return None

        shape = list(map(int, self.params.reshape.shape.value.split(",")))
        result = array.data.reshape(shape)

        # TODO: properly handle channel names
        del array.meta["channels"]
        return {"out": (result, array.meta)}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam


class ColorEnhancer(Node):
    def config_input_slots():
        return {"image": DataType.ARRAY}

    def config_output_slots():
        return {"enhanced_image": DataType.ARRAY}

    def config_params():
        return {
            "enhancement": {
                "contrast": FloatParam(1.0, 0.1, 10.0, doc="Multiplier for contrast adjustment. 1.0 means no change."),
                "brightness": FloatParam(0.0, -0.9, 0.9, doc="Additive brightness adjustment. 0.0 means no change."),
                "gamma": FloatParam(1.0, 0.1, 5.0, doc="Gamma correction value. 1.0 means no change."),
                "color_boost": FloatParam(1.0, 0.5, 3.0, doc="Multiplier for boosting colors. 1.0 means no change."),
            }
        }

    def process(self, image: Data):
        if image is None or image.data is None:
            return None

        # Extract parameters
        contrast = self.params["enhancement"]["contrast"].value
        brightness = self.params["enhancement"]["brightness"].value
        gamma = self.params["enhancement"]["gamma"].value
        color_boost = self.params["enhancement"]["color_boost"].value

        # Find the mean intensity
        midpoint = 0.5

        # Adjust contrast around the mean value
        enhanced_image = (image.data - midpoint) * contrast + midpoint + brightness

        # Adjust gamma
        enhanced_image = np.clip(np.power(enhanced_image, gamma), 0, 1)

        # Boost colors (only if the image is colored)
        if enhanced_image.shape[-1] == 3:
            enhanced_image *= color_boost
            enhanced_image = np.clip(enhanced_image, 0, 1)

        return {"enhanced_image": (enhanced_image, {**image.meta})}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node


class Correlation(Node):
    def config_input_slots():
        return {"data1": DataType.ARRAY, "data2": DataType.ARRAY}

    def config_output_slots():
        return {"pearson": DataType.ARRAY, "pval": DataType.ARRAY}

    def config_params():
        return {"correlation": {"axis": -1}}

    def setup(self):
        from scipy.stats import pearsonr

        self.pearsonr = pearsonr

    def process(self, data1: Data, data2: Data):
        if data1 is None or data2 is None:
            return None

        meta = data1.meta

        # broadcast data to same shape
        data1, data2 = np.broadcast_arrays(data1.data, data2.data)
        data = np.stack([data1, data2], axis=0)

        if data.ndim > 3:
            raise ValueError("Correlation only works for 1D and 2D data")

        # update axis param
        axis = self.params.correlation.axis.value
        if axis >= 0:
            axis += 1

        # calculate correlation along axis
        r, p = self.pearsonr(data[0], data[1], axis=axis)

        return {"pearson": (r, meta), "pval": (p, meta)}



import io

import numpy as np
import soundfile as sf

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, StringParam


class SpeechSynthesis(Node):
    def config_input_slots():
        return {"text": DataType.STRING, "voice": DataType.ARRAY}

    def config_output_slots():
        return {"speech": DataType.ARRAY, "transcript": DataType.STRING}

    def config_params():
        return {
            "speech_generation": {
                "openai_key": StringParam("openai.key"),
                "speed": FloatParam(1.0, 0.1, 2.0),
                "model": StringParam("tts-1", options=["tts-1", "tts-1-hd"]),
                "voice": StringParam("alloy", options=["alloy", "echo", "fable", "onyx", "nova", "shimmer"]),
            }
        }

    def setup(self):
        import librosa
        import openai

        self.librosa = librosa
        self.openai = openai

        key = self.params["speech_generation"]["openai_key"].value
        with open(key, "r") as f:
            key = f.read().strip()

        self.openai.api_key = key
        self.client = openai.OpenAI(api_key=key)

    def process(self, text: Data, voice: Data):
        speech = None
        transcript = ""

        if text is None and voice is None:
            return {"speech": (np.array([]), {}), "transcript": ("", {})}

        if text is not None:
            audio_generator = self.synthesize_speech_stream(text.data, self.params["speech_generation"]["speed"].value)
            # convert bytes to numpy array
            speech = self.bytes_to_array(b"".join([speech_bytes for speech_bytes in audio_generator]))
        elif voice is not None:
            # ensure transcript is a string
            transcript = self.transcribe_voice(voice.data) or ""

        return {"speech": (speech, {}), "transcript": (transcript, {})}

    def synthesize_speech_stream(self, text, speed):
        response = self.openai.audio.speech.create(
            model=self.params.speech_generation.model.value,
            voice=self.params.speech_generation.voice.value,
            input=text,
            speed=speed,
        )
        # yield audio chunks
        for chunk in response.iter_bytes():
            yield chunk

    def bytes_to_array(self, audio_bytes):
        with io.BytesIO(audio_bytes) as audio_file:
            with sf.SoundFile(audio_file) as sf_file:
                audio_array = np.array(sf_file.read(dtype="float32"))
                # resample if the sample rate is different from 44100 Hz
                if sf_file.samplerate != 44100:
                    audio_array = self.resample_audio(audio_array, sf_file.samplerate, 44100)
                return audio_array

    def resample_audio(self, audio, input_rate, output_rate):
        # resample audio from input_rate to output_rate
        return self.librosa.resample(audio, orig_sr=input_rate, target_sr=output_rate)

    def transcribe_voice(self, voice_buffer):
        # convert the numpy array buffer to a WAV file in memory
        with open("tmp.wav", "wb") as audio_stream:
            sf.write(audio_stream, voice_buffer, 44100, format="wav")

        with open("tmp.wav", "rb") as audio_stream:
            # send the audio file to OpenAI for transcription
            response = self.client.audio.transcriptions.create(model="whisper-1", file=audio_stream)

        return response.text



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, IntParam, StringParam


class Biotuner(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {
            "harmsim": DataType.ARRAY,
            "tenney": DataType.ARRAY,
            "subharm_tension": DataType.ARRAY,
            "cons": DataType.ARRAY,
            "peaks_ratios_tuning": DataType.ARRAY,
            "harm_tuning": DataType.ARRAY,
            "peaks": DataType.ARRAY,
            "amps": DataType.ARRAY,
            "extended_peaks": DataType.ARRAY,
            "extended_amps": DataType.ARRAY,
        }

    def config_params():
        return {
            "biotuner": {
                "n_peaks": IntParam(5, 1, 10, doc="Number of peaks to extract"),
                "f_min": FloatParam(2.0, 0.1, 50.0, doc="Minimum frequency"),
                "f_max": FloatParam(30.0, 1.0, 100.0, doc="Maximum frequency"),
                "precision": FloatParam(0.1, 0.01, 10.0, doc="Precision of the peak extraction in Hz"),
                "peaks_function": StringParam(
                    "EMD", options=["EMD", "fixed", "harmonic_recurrence", "EIMC"], doc="Peak extraction function"
                ),
                "n_harm_subharm": IntParam(3, 1, 10, doc="Number of harmonics to consider in the subharmonic tension metric"),
                "n_harm_extended": IntParam(3, 1, 10, doc="Number of harmonics to consider in the extended peaks"),
                "delta_lim": IntParam(250, 1, 300, doc="Maximum delta (in ms) in the subharmonic tension metric"),
            }
        }

    def process(self, data: Data):
        if data is None:
            return None

        data.data = np.squeeze(data.data)
        if data.data.ndim > 1:
            raise ValueError("Data must be 1D")

        result = biotuner_realtime(
            data.data,
            data.meta["sfreq"],
            n_peaks=self.params["biotuner"]["n_peaks"].value,
            peaks_function=self.params["biotuner"]["peaks_function"].value,
            min_freq=self.params["biotuner"]["f_min"].value,
            max_freq=self.params["biotuner"]["f_max"].value,
            precision=self.params["biotuner"]["precision"].value,
            n_harm_extended=self.params["biotuner"]["n_harm_extended"].value,
            n_harm_subharm=self.params["biotuner"]["n_harm_subharm"].value,
            delta_lim=self.params["biotuner"]["delta_lim"].value,
        )
        peaks, extended_peaks, metrics, tuning, harm_tuning, amps, extended_amps = result

        return {
            "harmsim": (np.array([metrics["harmsim"]]), data.meta),
            "tenney": (np.array([metrics["tenney"]]), data.meta),
            "subharm_tension": (np.array([metrics["subharm_tension"][0]]), data.meta),
            "cons": (np.array([metrics["cons"]]), data.meta),
            "peaks_ratios_tuning": (np.array(tuning), data.meta),
            "harm_tuning": (np.array(harm_tuning), data.meta),
            "peaks": (np.array(peaks), data.meta),
            "amps": (np.array(amps), data.meta),
            "extended_peaks": (np.array(extended_peaks), data.meta),
            "extended_amps": (np.array(extended_amps), data.meta),
        }


compute_biotuner_fn, harmonic_tuning_fn = None, None


def biotuner_realtime(
    data,
    sfreq,
    n_peaks=5,
    peaks_function="EMD",
    min_freq=1,
    max_freq=65,
    precision=0.1,
    n_harm_extended=3,
    n_harm_subharm=3,
    delta_lim=250,
):
    # import the biotuner function here to avoid loading it on startup
    global compute_biotuner_fn, harmonic_tuning_fn
    if compute_biotuner_fn is None or harmonic_tuning_fn is None:
        from biotuner.biotuner_object import compute_biotuner, harmonic_tuning

        compute_biotuner_fn = compute_biotuner
        harmonic_tuning_fn = harmonic_tuning

    # run biotuner peak extraction
    bt = compute_biotuner_fn(peaks_function=peaks_function, sf=sfreq)
    try:
        bt.peaks_extraction(
            np.array(data),
            graph=False,
            min_freq=min_freq,
            max_freq=max_freq,
            precision=precision,
            nIMFs=5,
            n_peaks=n_peaks,
            smooth_fft=1,
        )
    except UnboundLocalError:
        raise RuntimeError("No peaks found. Try increasing the length of the signal.")

    try:
        # try computing the extended peaks
        bt.peaks_extension(method="harmonic_fit", n_harm=n_harm_extended)
    except TypeError:
        raise RuntimeError("Detected only one peak. The largest peak might be below the minimum frequency.")

    bt.compute_peaks_metrics(n_harm=n_harm_subharm, delta_lim=delta_lim)
    if hasattr(bt, "all_harmonics"):
        harm_tuning = harmonic_tuning_fn(bt.all_harmonics)
    else:
        harm_tuning = [0, 1]
    peaks = bt.peaks
    amps = bt.amps
    extended_peaks = bt.extended_peaks
    extended_amps = bt.extended_amps
    metrics = bt.peaks_metrics

    if not isinstance(metrics["subharm_tension"][0], float):
        metrics["subharm_tension"] = [np.nan]
    metrics["harmsim"] = metrics["harmsim"] / 100
    # rescale tenney height from between 4 to 9 to between 0 and 1
    metrics["tenney"] = (metrics["tenney"] - 4) / 5
    tuning = bt.peaks_ratios
    return peaks, extended_peaks, metrics, tuning, harm_tuning, amps, extended_amps



import pytest

from goofi.manager import NodeContainer

from .utils import DummyNode


def test_creation():
    cont = NodeContainer()
    assert len(cont) == 0


def test_assignment():
    cont = NodeContainer()
    # the container shouldn't allow item assignment
    with pytest.raises(TypeError):
        cont["test"] = 1


def test_contains():
    cont = NodeContainer()
    assert "test" not in cont, "Empty container shouldn't contain anything"

    ref = DummyNode.create_local()[0]
    cont.add_node("test", ref)
    assert "test0" in cont, "Added node but container doesn't contain it"
    ref.terminate()


def test_add_node():
    cont = NodeContainer()

    # adding a node should increase the length of the container
    ref1 = DummyNode.create_local()[0]
    cont.add_node("test", ref1)
    assert len(cont) == 1, "Added node but length didn't increase"
    assert "test0" in cont, "Wrong name when adding node"

    # when adding a node with the same name, the name should be changed
    ref2 = DummyNode.create_local()[0]
    cont.add_node("test", ref2)
    assert len(cont) == 2, "Added node but length didn't increase"

    # make sure the name was changed
    assert "test1" in cont, "Wrong name when adding node"

    # check failure cases
    ref3 = DummyNode.create_local()[0]
    with pytest.raises(ValueError):
        cont.add_node(None, ref3)

    ref4 = DummyNode.create_local()[0]
    with pytest.raises(ValueError):
        cont.add_node(1, ref4)

    with pytest.raises(ValueError):
        cont.add_node("test", None)

    ref1.terminate()
    ref2.terminate()
    ref3.terminate()
    ref4.terminate()


def test_remove_node():
    cont = NodeContainer()
    cont.add_node("test", DummyNode.create_local()[0])

    # removing a node should decrease the length of the container
    cont.remove_node("test0")
    assert len(cont) == 0, "Removed node but length didn't decrease"

    # check failure cases
    with pytest.raises(KeyError):
        cont.remove_node("test0")
    with pytest.raises(KeyError):
        cont.remove_node(None)
    with pytest.raises(KeyError):
        cont.remove_node(1)



from abc import ABC, abstractmethod
from collections import namedtuple
from copy import deepcopy
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Union


@dataclass
class Param(ABC):
    """
    Parameter container that has a specific type and potentially constrains the range of allowed values.
    """

    _value: Any = None

    # NOTE: The `doc` and `save_param` attributes are added to the subclasses via monkey-patching.
    # The kw_only=True argument is not supported in Python<3.10, so we have to use this hacky solution.
    # doc: str = field(default=None, kw_only=True)
    # save_param: bool = field(default=True, kw_only=True)

    def __post_init__(self):
        if self._value is None:
            self._value = self.default()

        if not isinstance(self._value, type(self.default())):
            if isinstance(self.default(), float) and isinstance(self._value, int):
                # it's okay if we wanted a float but got int
                self._value = float(self._value)
            else:
                raise TypeError(
                    f"Parameter {self.__class__.__name__} expected type {type(self.default())} but got {type(self._value)}."
                )

    @staticmethod
    @abstractmethod
    def default() -> Any:
        """
        Return the default value for the parameter.
        """
        pass

    @property
    def value(self) -> Any:
        return self._value

    @value.setter
    def value(self, value: Any):
        self._value = value


@dataclass
class BoolParam(Param):
    trigger: bool = False

    @staticmethod
    def default() -> bool:
        return False

    @property
    def value(self) -> bool:
        """If trigger is True, return the current value and reset it to False."""
        val = self._value
        if self.trigger:
            self._value = False
        return val

    @value.setter
    def value(self, value: bool):
        self._value = value


@dataclass
class FloatParam(Param):
    vmin: float = 0.0
    vmax: float = 1.0

    @staticmethod
    def default() -> float:
        return 0.0


@dataclass
class IntParam(Param):
    vmin: int = -1
    vmax: int = 3

    @staticmethod
    def default() -> int:
        return 0


@dataclass
class StringParam(Param):
    # if options is None, the string is free-form, otherwise it is a dropdown
    options: List[str] = None

    @staticmethod
    def default() -> str:
        return ""


# NOTE: Monkey-patching the `doc` and `save_param` attribute into the Param subclasses is a hacky solution
# to include keyword-only arguments in the __init__ method of the subclasses. Python>=3.10 can use the
# kw_only=True argument in the field decorator.


# decorator for adjusting the __init__ method of the Param subclasses
def adjusted_init(original_init):
    def new_init(self, *args, **kwargs):
        self.doc = kwargs.pop("doc", None)
        self.save_param = kwargs.pop("save_param", True)
        original_init(self, *args, **kwargs)

    return new_init


# monkey-patch the 'doc' attribute into the classes
def add_extra_attributes(cls):
    # doc attribute
    setattr(cls, "doc", None)
    cls.__dataclass_fields__["doc"] = field(default=None)
    # save_param attribute
    setattr(cls, "save_param", True)
    cls.__dataclass_fields__["save_param"] = field(default=True)

    # adjust the __init__ method
    cls.__init__ = adjusted_init(cls.__init__)


# apply the monkey-patching
add_extra_attributes(Param)
for subclass in Param.__subclasses__():
    add_extra_attributes(subclass)


DEFAULT_PARAMS = {
    "common": {
        "autotrigger": BoolParam(
            False,
            doc=(
                "If set, the node will run its processing function automatically "
                "at the specified Max Frequency, even if none of the inputs have changed."
            ),
        ),
        "max_frequency": FloatParam(
            30.0, 0.0, 60.0, doc="The maximum frequency at which the node will run its processing function."
        ),
    },
}

TYPE_PARAM_MAP = {
    bool: BoolParam,
    float: FloatParam,
    int: IntParam,
    str: StringParam,
}


class InvalidParamError(Exception):
    pass


class NodeParams:
    """
    A class for storing node parameters, which store the configuration state of the node, and are exposed
    to the user through the GUI. The parameters are stored in named groups with each group containing a
    number of parameters. The parameters are stored as named tuples, and can be accessed as attributes.

    When initializing a `NodeParams` object, a set of default parameters is inserted if they are not provided.

    ### Parameters
    `data` : Dict[str, Dict[str, Any]]
        A dictionary of parameter groups, where each group is a dictionary of parameter names and values.
    """

    def __init__(self, data: Dict[str, Dict[str, Any]]):
        # insert default parameters if they are not present
        for group_name, group in deepcopy(DEFAULT_PARAMS).items():
            if group_name not in data:
                # group is not present, insert it
                data[group_name] = group
            else:
                # group is present, insert missing parameters
                for param_name, param in group.items():
                    if param_name in data[group_name]:
                        # make sure we have a Param and not just a deserialized value
                        if not isinstance(data[group_name][param_name], Param):
                            param._value = data[group_name][param_name]
                            data[group_name][param_name] = param
                    else:
                        data[group_name][param_name] = param

        # convert values to Param objects
        for group, params in data.items():
            if not isinstance(params, dict):
                raise TypeError(f"Expected dict, got {type(params)}.")
            for param_name, param in params.items():
                if not isinstance(param, Param):
                    if isinstance(param, dict):
                        # reconstruct serialized param object
                        param_type = TYPE_PARAM_MAP[type(param["_value"])]
                        data[group][param_name] = param_type(**param)
                        continue

                    # convert to Param object
                    if type(param) not in TYPE_PARAM_MAP:
                        raise TypeError(
                            f"Invalid parameter type {type(param).__name__}. Must be one of "
                            f"{list(map(lambda x: x.__name__, TYPE_PARAM_MAP.keys()))}."
                        )
                    data[group][param_name] = TYPE_PARAM_MAP[type(param)](param)

        # convert to named tuples
        self._data = self._generate_data_dict(data)

    def _generate_data_dict(self, data: Dict[str, Dict[str, Any]]) -> Dict[str, Dict[str, Any]]:
        """
        Convert the parameter dictionary to a dictionary of named tuples.

        ### Parameters
        `data` : Dict[str, Dict[str, Any]]
            A dictionary of parameter groups, where each group is a dictionary of parameter names and values.

        ### Returns
        `Dict[str, Dict[str, Any]]`
            A dictionary of parameter groups, where each group is a dictionary of parameter names and values.
        """
        # convert to named tuples
        result = {}
        for group, params in data.items():
            # create the named tuple class for the current group
            NamedTupleClass = namedtuple(group.capitalize(), params.keys())

            # implement __contains__ for the named tuple class
            NamedTupleClass = type(
                NamedTupleClass.__name__,
                (NamedTupleClass,),
                {
                    "__contains__": lambda self, item: hasattr(self, item),
                    "__getitem__": lambda self, item: getattr(self, item),
                    "keys": lambda self: self._asdict().keys(),
                    "values": lambda self: self._asdict().values(),
                    "items": lambda self: self._asdict().items(),
                },
            )

            result[group] = NamedTupleClass(**params)
        return result

    def update(self, params: Dict[str, Dict[str, Any]]):
        """
        Update the parameters with new values.

        ### Parameters
        `params` : Dict[str, Dict[str, Any]]
            A dictionary of parameter groups, where each group is a dictionary of parameter names and values.
        """
        encountered_invalid_params = False

        # TODO: avoid code duplication with __init__
        for group, params in params.items():
            if group not in self._data:
                encountered_invalid_params = True
                print(f"Invalid parameter group '{group}'.")
                continue

            for name, param in params.items():
                if name not in self._data[group]._fields:
                    encountered_invalid_params = True
                    print(f"Invalid parameter '{name}' in group '{group}'.")
                    continue

                if not isinstance(param, Param):
                    if isinstance(param, dict):
                        # !!! LOADING PARAMS FROM DICT IS A LEGACY FEATURE TO LOAD OLD GOOFI PATCHES !!!
                        # reconstruct serialized param object
                        param_type = type(self._data[group][name])
                        self._data[group] = self._data[group]._replace(**{name: param_type(**param)})
                        continue

                    # update only the value of the parameter, leave the rest unchanged
                    if not isinstance(self._data[group][name], TYPE_PARAM_MAP[type(param)]):
                        # it's okay if we wanted a float but got int
                        if isinstance(self._data[group][name], FloatParam) and isinstance(param, int):
                            param = float(param)
                        else:
                            raise TypeError(
                                f"Expected parameter type {TYPE_PARAM_MAP[type(param)].__name__} but got "
                                f"{type(self._data[group][name]).__name__}."
                            )
                    self._data[group][name]._value = param
                else:
                    # update the entire parameter object
                    self._data[group] = self._data[group]._replace(**{name: param})

        if encountered_invalid_params:
            raise InvalidParamError("Invalid parameters encountered.")

    def serialize(self) -> Dict[str, Dict[str, Any]]:
        """
        Serialize the parameter values to a dictionary.

        ### Returns
        `Dict[str, Dict[str, Any]]`
            A dictionary of parameter groups, where each group is a dictionary of parameter names and values.
        """
        serialized_data = {}
        for group, params in self._data.items():
            serialized_params = {}
            for name, param in params.items():
                if param.save_param:
                    param = param._value
                    serialized_params[name] = param
            serialized_data[group] = serialized_params
        return serialized_data

    def __getattr__(self, group: str):
        # don't allow access to the _data attribute
        if group == "_data":
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{group}'")

        # return the group if it exists
        if group in self._data:
            return self._data[group]
        else:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{group}'")

    def __contains__(self, group: str) -> bool:
        return group in self._data

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NodeParams):
            return False
        return self._data == other._data

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(map(repr,self._data.values()))})"

    def __getitem__(self, group: Union[int, str]):
        if isinstance(group, int):
            return list(self._data.keys())[group]
        return self._data[group]

    def __len__(self) -> int:
        return len(self._data)

    def __getstate__(self):
        return {"_data": {k: v._asdict() for k, v in self._data.items()}}

    def __setstate__(self, state):
        self._data = self._generate_data_dict(state["_data"])
